from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_reddit
from arcade_reddit.enums import RedditTimeFilter, SubredditListingType
from arcade_reddit.tools import (
    get_content_of_post,
    get_posts_in_subreddit,
    get_top_level_comments,
)
from arcade_reddit.tools.read import (
    check_subreddit_access,
    get_content_of_multiple_posts,
    get_my_posts,
    get_subreddit_rules,
)
from evals.additional_messages import get_post_in_subreddit_messages
from evals.critics import AnyOfCritic, ListCritic

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_reddit)


@tool_eval()
def reddit_get_posts_in_subreddit_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_get_posts_in_subreddit_1",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_get_posts_in_subreddit_1",
        user_message="Get 30 posts from AskReddit that are contentious at this moment.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_posts_in_subreddit,
                args={
                    "subreddit": "AskReddit",
                    "listing": SubredditListingType.CONTROVERSIAL.value,
                    "limit": 30,
                    "cursor": None,
                    "time_range": RedditTimeFilter.NOW.value,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=0.3),
            BinaryCritic(critic_field="listing", weight=0.2),
            BinaryCritic(critic_field="limit", weight=0.2),
            BinaryCritic(critic_field="cursor", weight=0.1),
            BinaryCritic(critic_field="time_range", weight=0.2),
        ],
    )

    suite.add_case(
        name="reddit_get_posts_in_subreddit_2",
        user_message="Get the next 5 posts from AskReddit that are contentious at this moment.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_posts_in_subreddit,
                args={
                    "subreddit": "AskReddit",
                    "listing": SubredditListingType.CONTROVERSIAL.value,
                    "limit": 5,
                    "cursor": "t3_1abcdef",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=0.2),
            BinaryCritic(critic_field="listing", weight=0.2),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="cursor", weight=0.3),
            BinaryCritic(critic_field="time_range", weight=0.2),
        ],
        additional_messages=get_post_in_subreddit_messages,
    )

    suite.add_case(  # time-based listing, but don't provide a specific time range
        name="reddit_get_posts_in_subreddit_3",
        user_message="Get 5 top posts from AskReddit",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_posts_in_subreddit,
                args={
                    "subreddit": "AskReddit",
                    "listing": SubredditListingType.TOP.value,
                    "limit": 5,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=0.3),
            BinaryCritic(critic_field="listing", weight=0.3),
            BinaryCritic(critic_field="limit", weight=0.3),
        ],
    )

    suite.add_case(
        name="reddit_get_posts_in_subreddit_4",
        user_message="Get posts from AskReddit that are gaining traction as we speak",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_posts_in_subreddit,
                args={
                    "subreddit": "AskReddit",
                    "listing": SubredditListingType.RISING.value,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=0.3),
            BinaryCritic(critic_field="listing", weight=0.7),
        ],
    )

    return suite


@tool_eval()
def reddit_get_content_of_post_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_get_content_of_post",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_get_content_of_post_1",
        user_message="Get the content of the post with the id t3_1abcdef",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_content_of_post,
                args={
                    "post_identifier": [  # post_identifier can be any of the following
                        "1abcdef",
                        "t3_1abcdef",
                        "https://www.reddit.com/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/",
                        "/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/",
                    ],
                },
            ),
        ],
        rubric=rubric,
        critics=[
            AnyOfCritic(
                critic_field="post_identifier",
                weight=1.0,
            ),
        ],
        additional_messages=get_post_in_subreddit_messages,
    )

    return suite


@tool_eval()
def reddit_get_content_of_multiple_posts_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_get_content_of_multiple_posts",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_get_content_of_multiple_posts_1",
        user_message=(
            "Get the content of the posts t3_1abcdef, "
            "https://www.reddit.com/r/AskReddit/comments/1jdfgk1vn/why_is_water_wet/, "
            "t3_3abcdef, t3_4abcdef, and t3_5abcdef, t3_6abcdef, and t3_7abcdef, 4jfnsklf, "
            "/r/AskReddit/comments/2dsghr/, and /r/AskReddit/comments/1jdfg35dvn/"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_content_of_multiple_posts,
                args={
                    "post_identifiers": [
                        "t3_1abcdef",
                        "https://www.reddit.com/r/AskReddit/comments/1jdfgk1vn/why_is_water_wet/",
                        "t3_3abcdef",
                        "t3_4abcdef",
                        "t3_5abcdef",
                        "t3_6abcdef",
                        "t3_7abcdef",
                        "4jfnsklf",
                        "/r/AskReddit/comments/2dsghr/",
                        "/r/AskReddit/comments/1jdfg35dvn/",
                    ],
                },
            ),
        ],
        rubric=rubric,
        critics=[
            ListCritic(
                critic_field="post_identifiers",
                weight=1.0,
                order_matters=False,
                duplicates_matter=True,
            ),
        ],
    )

    return suite


@tool_eval()
def reddit_get_top_level_comments_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_get_top_level_comments",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_get_top_level_comments_1",
        user_message="What are people saying in response?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_top_level_comments,
                args={
                    "post_identifier": [  # post_identifier can be any of the following
                        "1abcdef",
                        "t3_1abcdef",
                        "https://www.reddit.com/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/",
                        "/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/",
                    ],
                },
            ),
        ],
        rubric=rubric,
        critics=[
            AnyOfCritic(
                critic_field="post_identifier",
                weight=1.0,
            ),
        ],
        additional_messages=get_post_in_subreddit_messages,
    )

    return suite


@tool_eval()
def reddit_check_subreddit_access_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_check_subreddit_access",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_check_subreddit_access_1",
        user_message="does r/WaterBottleCollecting exist?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=check_subreddit_access,
                args={
                    "subreddit": "WaterBottleCollecting",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=1.0),
        ],
    )

    suite.add_case(
        name="reddit_check_subreddit_access_2",
        user_message=(
            "so my friend is a part of the WaterBottleCollecting subreddit, "
            "but i cant find it. Why?"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=check_subreddit_access,
                args={
                    "subreddit": "WaterBottleCollecting",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=1.0),
        ],
    )

    return suite


@tool_eval()
def reddit_get_subreddit_rules_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_get_subreddit_rules",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_get_subreddit_rules_1",
        user_message=(
            "I'm going to be posting some stuff on WaterBottleCollecting, "
            "but I'm scared that I might go against their terms & conditions "
            "and get my post removed."
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_subreddit_rules,
                args={"subreddit": "WaterBottleCollecting"},
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=1.0),
        ],
    )

    suite.add_case(
        name="reddit_get_subreddit_rules_2",
        user_message=(
            "What are WaterBottleCollecting's bannable offenses? I don't want to get banned!"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_subreddit_rules,
                args={"subreddit": "WaterBottleCollecting"},
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=1.0),
        ],
    )

    return suite


@tool_eval()
def reddit_get_my_posts_eval_suite() -> EvalSuite:
    get_my_posts_response = [
        {"role": "user", "content": "get 1 of my posts"},
        {
            "role": "assistant",
            "content": "",
            "tool_calls": [
                {
                    "id": "call_hPacHNSvuKKamoKKoPWBQosv",
                    "type": "function",
                    "function": {"name": "Reddit_GetMyPosts", "arguments": '{"limit":1}'},
                }
            ],
        },
        {
            "role": "tool",
            "content": '{"cursor":"t3_1jt9jz4","posts":[{"author":"RedditUser123","body":"i just wanted to say that i love thisapp\\n\\n","created_utc":1743988489,"id":"1jt9jz4","is_video":false,"name":"t3_1jt9jz4","num_comments":1,"permalink":"/r/SparkingWater/comments/1jt9jz4/this_is_fun/","score":1,"subreddit":"SparklingWater","title":"this isfun","upvote_ratio":1,"upvotes":1,"url":"https://www.reddit.com/r/SparklingWater/comments/1jt9jz4/this_is_fun/"}]}',  # noqa: E501
            "tool_call_id": "call_hPacHNSvuKKamoKKoPWBQosv",
            "name": "Reddit_GetMyPosts",
        },
        {
            "role": "assistant",
            "content": "Here is one of your posts on Reddit:\n\n**Title:** [this isfun](https://www.reddit.com/r/SparklingWater/comments/1jt9jz4/this_is_fun/)\n\n**Subreddit:**r/SparklingWater\n\n**Content:** \n```\ni just wanted to say that i love this app\n```\n\n**Upvotes:** 1  \n**Comments:** 1",  # noqa: E501
        },
    ]

    suite = EvalSuite(
        name="reddit_get_my_posts",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_get_my_posts_1",
        user_message=(
            "I want to train an AI on the voice I use for my reddit posts. "
            "Help me out here & get my last 100"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_my_posts,
                args={
                    "limit": 100,
                    "include_body": True,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="include_body", weight=0.5),
        ],
    )

    suite.add_case(
        name="reddit_get_my_posts_2",
        user_message=("get 25 more but w/o their content"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_my_posts,
                args={
                    "limit": 25,
                    "include_body": False,
                    "cursor": "t3_1jt9jz4",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.3),
            BinaryCritic(critic_field="include_body", weight=0.3),
            BinaryCritic(critic_field="cursor", weight=0.4),
        ],
        additional_messages=get_my_posts_response,
    )

    return suite



from enum import Enum
from typing import cast


class SalesforceObject(Enum):
    ACCOUNT = "Account"
    CALL = "Call"
    CONTACT = "Contact"
    EMAIL = "Email"
    EVENT = "Event"
    LEAD = "Lead"
    NOTE = "Note"
    OPPORTUNITY = "Opportunity"
    TASK = "Task"
    USER = "User"

    @property
    def plural(self) -> str:
        if self == SalesforceObject.OPPORTUNITY:
            return "Opportunities"
        return cast(str, self.value) + "s"



from typing import Annotated

import httpx
from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import GitHub

from arcade_github.tools.utils import get_github_json_headers, get_url, handle_github_response


# Implements https://docs.github.com/en/rest/activity/starring?apiVersion=2022-11-28#star-a-repository-for-the-authenticated-user and https://docs.github.com/en/rest/activity/starring?apiVersion=2022-11-28#unstar-a-repository-for-the-authenticated-user  # noqa: E501
# Example `arcade chat` usage: "star the vscode repo owned by microsoft"
@tool(requires_auth=GitHub())
async def set_starred(
    context: ToolContext,
    owner: Annotated[str, "The owner of the repository"],
    name: Annotated[str, "The name of the repository"],
    starred: Annotated[bool, "Whether to star the repository or not"] = True,
) -> Annotated[
    str, "A message indicating whether the repository was successfully starred or unstarred"
]:
    """
    Star or un-star a GitHub repository.
    For example, to star microsoft/vscode, you would use:
    ```
    set_starred(owner="microsoft", name="vscode", starred=True)
    ```
    """
    url = get_url("user_starred", owner=owner, repo=name)
    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        if starred:
            response = await client.put(url, headers=headers)
        else:
            response = await client.delete(url, headers=headers)

    handle_github_response(response, url)

    action = "starred" if starred else "unstarred"
    return f"Successfully {action} the repository {owner}/{name}"


# Implements https://docs.github.com/en/rest/activity/starring?apiVersion=2022-11-28#list-stargazers
# Example `arcade chat` usage: "list the stargazers for the ArcadeAI/arcade-ai repo"
@tool(requires_auth=GitHub())
async def list_stargazers(
    context: ToolContext,
    owner: Annotated[str, "The owner of the repository"],
    repo: Annotated[str, "The name of the repository"],
    limit: Annotated[
        int | None,
        "The maximum number of stargazers to return. "
        "If not provided, all stargazers will be returned.",
    ] = None,
) -> Annotated[dict, "A dictionary containing the stargazers for the specified repository"]:
    """List the stargazers for a GitHub repository."""
    url = get_url("repo_stargazers", owner=owner, repo=repo)
    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    if limit is None:
        limit = 2**64 - 1

    per_page = min(limit, 100)
    page = 1
    stargazers: list[dict] = []

    async with httpx.AsyncClient() as client:
        while len(stargazers) < limit:
            response = await client.get(
                url, headers=headers, params={"per_page": per_page, "page": page}
            )
            handle_github_response(response, url)

            data = response.json()
            if not data:
                break

            stargazers.extend([
                {
                    "login": stargazer.get("login"),
                    "id": stargazer.get("id"),
                    "node_id": stargazer.get("node_id"),
                    "html_url": stargazer.get("html_url"),
                }
                for stargazer in data
            ])

            if len(data) < per_page:
                break

            page += 1

    stargazers = stargazers[:limit]
    return {"number_of_stargazers": len(stargazers), "stargazers": stargazers}



import logging
from typing import Any

from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource, build

from arcade_google_drive.enums import Corpora, OrderBy

## Set up basic configuration for logging to the console with DEBUG level and a specific format.
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)


def build_drive_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build a Drive service object.
    """
    auth_token = auth_token or ""
    return build("drive", "v3", credentials=Credentials(auth_token))


def build_file_tree_request_params(
    order_by: list[OrderBy] | None,
    page_token: str | None,
    limit: int | None,
    include_shared_drives: bool,
    restrict_to_shared_drive_id: str | None,
    include_organization_domain_documents: bool,
) -> dict[str, Any]:
    if order_by is None:
        order_by = [OrderBy.MODIFIED_TIME_DESC]
    elif isinstance(order_by, OrderBy):
        order_by = [order_by]

    params = {
        "q": "trashed = false",
        "corpora": Corpora.USER.value,
        "pageToken": page_token,
        "fields": (
            "files(id, name, parents, mimeType, driveId, size, createdTime, modifiedTime, owners)"
        ),
        "orderBy": ",".join([item.value for item in order_by]),
    }

    if limit:
        params["pageSize"] = str(limit)

    if (
        include_shared_drives
        or restrict_to_shared_drive_id
        or include_organization_domain_documents
    ):
        params["includeItemsFromAllDrives"] = "true"
        params["supportsAllDrives"] = "true"

    if restrict_to_shared_drive_id:
        params["driveId"] = restrict_to_shared_drive_id
        params["corpora"] = Corpora.DRIVE.value

    if include_organization_domain_documents:
        params["corpora"] = Corpora.DOMAIN.value

    return params


def build_file_tree(files: dict[str, Any]) -> dict[str, Any]:
    file_tree: dict[str, Any] = {}

    for file in files.values():
        owners = file.get("owners", [])
        if owners:
            owners = [
                {"name": owner.get("displayName", ""), "email": owner.get("emailAddress", "")}
                for owner in owners
            ]
            file["owners"] = owners

        if "size" in file:
            file["size"] = {"value": int(file["size"]), "unit": "bytes"}

        # Although "parents" is a list, a file can only have one parent
        try:
            parent_id = file["parents"][0]
            del file["parents"]
        except (KeyError, IndexError):
            parent_id = None

        # Determine the file's Drive ID
        if "driveId" in file:
            drive_id = file["driveId"]
            del file["driveId"]
        # If a shared drive id is not present, the file is in "My Drive"
        else:
            drive_id = "My Drive"

        if drive_id not in file_tree:
            file_tree[drive_id] = []

        # Root files will have the Drive's id as the parent. If the parent id is not in the files
        # list, the file must be at drive's root
        if parent_id not in files:
            file_tree[drive_id].append(file)

        # Associate the file with its parent
        else:
            if "children" not in files[parent_id]:
                files[parent_id]["children"] = []
            files[parent_id]["children"].append(file)

    return file_tree



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_dropbox
from arcade_dropbox.critics import DropboxPathCritic
from arcade_dropbox.tools.files import download_file

rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
catalog.add_module(arcade_dropbox)


@tool_eval()
def download_file_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the download_file tool."""
    suite = EvalSuite(
        name="download_file",
        system_message="You are an AI assistant that can interact with files and folders in Dropbox using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Download file in the root folder by file path",
        user_message="Download the file test.txt from Dropbox",
        expected_tool_calls=[
            ExpectedToolCall(
                func=download_file,
                args={
                    "file_path": "test.txt",
                    "file_id": None,
                },
            ),
        ],
        critics=[
            DropboxPathCritic(critic_field="file_path", weight=0.5),
            BinaryCritic(critic_field="file_id", weight=0.5),
        ],
    )

    suite.add_case(
        name="Download file with a sub-folder structure",
        user_message="Download the file Q1report.ppt in the folder AcmeInc/Reports from Dropbox",
        expected_tool_calls=[
            ExpectedToolCall(
                func=download_file,
                args={
                    "file_path": "/AcmeInc/Reports/Q1report.ppt",
                    "file_id": None,
                },
            ),
        ],
        critics=[
            DropboxPathCritic(critic_field="file_path", weight=0.5),
            BinaryCritic(critic_field="file_id", weight=0.5),
        ],
    )

    suite.add_case(
        name="Download file by ID",
        user_message="Download the file id:a4ayc_80_OEAAAAAAAAAYa from Dropbox",
        expected_tool_calls=[
            ExpectedToolCall(
                func=download_file,
                args={
                    "file_path": None,
                    "file_id": "id:a4ayc_80_OEAAAAAAAAAYa",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="file_path", weight=0.5),
            BinaryCritic(critic_field="file_id", weight=0.5),
        ],
    )

    return suite



from typing import TYPE_CHECKING, Any

from arcade_core.schema import ToolDefinition
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

if TYPE_CHECKING:
    from arcade_evals.eval import EvaluationResult
console = Console()


def display_tools_table(tools: list[ToolDefinition]) -> None:
    """
    Display a table of tools with their name, description, package, and version.
    """
    if not tools:
        console.print("No tools found.", style="bold")
        return

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Name")
    table.add_column("Description")
    table.add_column("Package")
    table.add_column("Version")

    for tool in sorted(tools, key=lambda x: x.toolkit.name):
        table.add_row(
            str(tool.get_fully_qualified_name()),
            tool.description.split("\n")[0] if tool.description else "",
            tool.toolkit.name,
            tool.toolkit.version,
        )
    console.print(f"Found {len(tools)} tools.")
    console.print(table)


def display_tool_details(tool: ToolDefinition) -> None:
    """
    Display detailed information about a specific tool using multiple panels.
    """
    # Description Panel
    description_panel = Panel(
        tool.description or "No description available.",
        title=f"Tool: {tool.name}",
        border_style="cyan",
    )

    # Inputs Panel
    inputs = tool.input.parameters
    if inputs:
        inputs_table = Table(show_header=True, header_style="bold green")
        inputs_table.add_column("Name", style="cyan")
        inputs_table.add_column("Type", style="magenta")
        inputs_table.add_column("Required", style="yellow")
        inputs_table.add_column("Description", style="white")
        inputs_table.add_column("Default", style="blue")
        for param in inputs:
            # Since InputParameter does not have a default field, we use "N/A"
            default_value = "N/A"
            if param.value_schema.enum:
                default_value = f"One of {param.value_schema.enum}"
            inputs_table.add_row(
                param.name,
                param.value_schema.val_type,
                str(param.required),
                param.description or "",
                default_value,
            )
        inputs_panel = Panel(
            inputs_table,
            title="Input Parameters",
            border_style="green",
        )
    else:
        inputs_panel = Panel(
            "No input parameters.",
            title="Input Parameters",
            border_style="green",
        )

    # Output Panel
    output = tool.output
    if output:
        output_description = output.description or "No description available."
        output_types = ", ".join(output.available_modes)
        output_val_type = output.value_schema.val_type if output.value_schema else "N/A"
        output_details = Text.assemble(
            ("Description: ", "bold"),
            (output_description, ""),
            "\n",
            ("Available Modes: ", "bold"),
            (output_types, ""),
            "\n",
            ("Value Type: ", "bold"),
            (output_val_type, ""),
        )
        output_panel = Panel(
            output_details,
            title="Expected Output",
            border_style="blue",
        )
    else:
        output_panel = Panel(
            "No output information available.",
            title="Expected Output",
            border_style="blue",
        )

    # Combine all panels vertically
    console.print(description_panel)
    console.print(inputs_panel)
    console.print(output_panel)


def display_tool_messages(tool_messages: list[dict]) -> None:
    for message in tool_messages:
        if message["role"] == "assistant":
            for tool_call in message.get("tool_calls", []):
                console.print(
                    f"[bold]Called tool '{tool_call['function']['name']}' with parameters:[/bold] {tool_call['function']['arguments']}",
                    style="dim",
                )
        elif message["role"] == "tool":
            console.print(
                f"[bold]'{message['name']}' tool returned:[/bold] {message['content']}", style="dim"
            )


def display_eval_results(results: list[list[dict[str, Any]]], show_details: bool = False) -> None:
    """
    Display evaluation results in a format inspired by pytest's output.

    Args:
        results: List of dictionaries containing evaluation results for each model.
        show_details: Whether to show detailed results for each case.
    """
    total_passed = 0
    total_failed = 0
    total_warned = 0
    total_cases = 0

    for eval_suite in results:
        for model_results in eval_suite:
            model = model_results.get("model", "Unknown Model")
            rubric = model_results.get("rubric", "Unknown Rubric")
            cases = model_results.get("cases", [])
            total_cases += len(cases)

            console.print(f"[bold]Model:[/bold] [bold magenta]{model}[/bold magenta]")
            if show_details:
                console.print(f"[bold magenta]{rubric}[/bold magenta]")

            for case in cases:
                evaluation = case["evaluation"]
                status = (
                    "[green]PASSED[/green]"
                    if evaluation.passed
                    else "[yellow]WARNED[/yellow]"
                    if evaluation.warning
                    else "[red]FAILED[/red]"
                )
                if evaluation.passed:
                    total_passed += 1
                elif evaluation.warning:
                    total_warned += 1
                else:
                    total_failed += 1

                # Display one-line summary for each case with score as a percentage
                score_percentage = evaluation.score * 100
                console.print(f"{status} {case['name']} -- Score: {score_percentage:.2f}%")

                if show_details:
                    # Show detailed information for each case
                    console.print(f"[bold]User Input:[/bold] {case['input']}\n")
                    console.print("[bold]Details:[/bold]")
                    console.print(_format_evaluation(evaluation))
                    console.print("-" * 80)

    # Summary
    summary = (
        f"[bold]Summary -- [/bold]Total: {total_cases} -- [green]Passed: {total_passed}[/green]"
    )
    if total_warned > 0:
        summary += f" -- [yellow]Warnings: {total_warned}[/yellow]"
    if total_failed > 0:
        summary += f" -- [red]Failed: {total_failed}[/red]"
    console.print(summary + "\n")


def _format_evaluation(evaluation: "EvaluationResult") -> str:
    """
    Format evaluation results with color-coded matches and scores.

    Args:
        evaluation: An EvaluationResult object containing the evaluation results.

    Returns:
        A formatted string representation of the evaluation details.
    """
    result_lines = []
    if evaluation.failure_reason:
        result_lines.append(f"[bold red]Failure Reason:[/bold red] {evaluation.failure_reason}")
    else:
        for critic_result in evaluation.results:
            is_criticized = critic_result.get("is_criticized", True)
            match_color = (
                "yellow" if not is_criticized else "green" if critic_result["match"] else "red"
            )
            field = critic_result["field"]
            score = critic_result["score"]
            weight = critic_result["weight"]
            expected = critic_result["expected"]
            actual = critic_result["actual"]

            if is_criticized:
                result_lines.append(
                    f"[bold]{field}:[/bold] "
                    f"[{match_color}]Match: {critic_result['match']}"
                    f"\n     Score: {score:.2f}/{weight:.2f}[/{match_color}]"
                    f"\n     Expected: {expected}"
                    f"\n     Actual: {actual}"
                )
            else:
                result_lines.append(
                    f"[bold]{field}:[/bold] "
                    f"[{match_color}]Un-criticized[/{match_color}]"
                    f"\n     Expected: {expected}"
                    f"\n     Actual: {actual}"
                )
    return "\n".join(result_lines)


def display_arcade_chat_header(base_url: str, stream: bool) -> None:
    chat_header = Text.assemble(
        "\n",
        (
            "=== Arcade Chat ===",
            "bold magenta underline",
        ),
        "\n",
        "\n",
        "Chatting with Arcade Engine at ",
        (
            base_url,
            "bold blue",
        ),
    )
    if stream:
        chat_header.append(" (streaming)")
    console.print(chat_header)



from unittest.mock import MagicMock

import httpx
import pytest

from arcade_dropbox.tools.browse import list_items_in_folder
from arcade_dropbox.utils import clean_dropbox_entries


@pytest.mark.asyncio
async def test_list_items_success_empty_folder(
    mock_context,
    mock_httpx_client,
):
    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"entries": [], "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await list_items_in_folder(
        context=mock_context,
        folder_path="/path/to/folder",
    )

    assert tool_response == {
        "items": [],
        "cursor": None,
        "has_more": False,
    }


@pytest.mark.asyncio
async def test_list_items_success_with_folder_entries(
    mock_context,
    mock_httpx_client,
    sample_folder_entry,
    sample_file_entry,
):
    entries = [sample_folder_entry, sample_file_entry]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"entries": entries, "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await list_items_in_folder(
        context=mock_context,
        folder_path="/path/to/folder",
    )

    assert tool_response == {
        "items": clean_dropbox_entries(entries),
        "cursor": None,
        "has_more": False,
    }


@pytest.mark.asyncio
async def test_list_items_success_with_more_items_to_paginate(
    mock_context,
    mock_httpx_client,
    sample_folder_entry,
    sample_file_entry,
):
    entries = [sample_folder_entry, sample_file_entry]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {
        "entries": entries,
        "cursor": "cursor",
        "has_more": True,
    }
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await list_items_in_folder(
        context=mock_context,
        folder_path="/path/to/folder",
    )

    assert tool_response == {
        "items": clean_dropbox_entries(entries),
        "cursor": "cursor",
        "has_more": True,
    }


@pytest.mark.asyncio
async def test_list_items_success_providing_cursor(
    mock_context,
    mock_httpx_client,
    sample_folder_entry,
    sample_file_entry,
):
    entries = [sample_folder_entry, sample_file_entry]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {
        "entries": entries,
        "cursor": "cursor2",
        "has_more": True,
    }
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await list_items_in_folder(
        context=mock_context,
        folder_path="/path/to/folder",
        cursor="cursor1",
        limit=2,
    )

    assert tool_response == {
        "items": clean_dropbox_entries(entries),
        "cursor": "cursor2",
        "has_more": True,
    }

    # Check that the request was made with the cursor and not the other arguments
    mock_httpx_client.post.assert_called_with(
        url="https://api.dropboxapi.com/2/files/list_folder/continue",
        headers={"Authorization": "Bearer fake-token"},
        json={"cursor": "cursor1"},
    )


@pytest.mark.asyncio
async def test_list_items_path_not_found(
    mock_context,
    mock_httpx_client,
):
    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 409
    mock_httpx_response.json.return_value = {"error_summary": "path/not_found"}

    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await list_items_in_folder(
        context=mock_context,
        folder_path="/not/exist/folder",
    )

    assert tool_response == {
        "error": "The specified path was not found by Dropbox",
    }



import asyncio
import json
from dataclasses import dataclass
from typing import Optional, cast

import httpx

from arcade_asana.constants import ASANA_API_VERSION, ASANA_BASE_URL, ASANA_MAX_CONCURRENT_REQUESTS
from arcade_asana.decorators import clean_asana_response
from arcade_asana.exceptions import AsanaToolExecutionError


@dataclass
class AsanaClient:
    auth_token: str
    base_url: str = ASANA_BASE_URL
    api_version: str = ASANA_API_VERSION
    max_concurrent_requests: int = ASANA_MAX_CONCURRENT_REQUESTS
    _semaphore: asyncio.Semaphore | None = None

    def __post_init__(self) -> None:
        self._semaphore = self._semaphore or asyncio.Semaphore(self.max_concurrent_requests)

    def _build_url(self, endpoint: str, api_version: str | None = None) -> str:
        api_version = api_version or self.api_version
        return f"{self.base_url.rstrip('/')}/{api_version.strip('/')}/{endpoint.lstrip('/')}"

    def _build_error_messages(self, response: httpx.Response) -> tuple[str, str]:
        try:
            data = response.json()
            errors = data["errors"]

            if len(errors) == 1:
                error_message = errors[0]["message"]
                developer_message = (
                    f"{errors[0]['message']} | {errors[0]['help']} "
                    f"(HTTP status code: {response.status_code})"
                )
            else:
                errors_concat = "', '".join([error["message"] for error in errors])
                error_message = f"Multiple errors occurred: '{errors_concat}'"
                developer_message = (
                    f"Multiple errors occurred: {json.dumps(errors)} "
                    f"(HTTP status code: {response.status_code})"
                )

        except Exception as e:
            error_message = "Failed to parse Asana error response"
            developer_message = f"Failed to parse Asana error response: {type(e).__name__}: {e!s}"

        return error_message, developer_message

    def _raise_for_status(self, response: httpx.Response) -> None:
        if response.status_code < 300:
            return

        error_message, developer_message = self._build_error_messages(response)

        raise AsanaToolExecutionError(error_message, developer_message)

    def _set_request_body(self, kwargs: dict, data: dict | None, json_data: dict | None) -> dict:
        if data and json_data:
            raise ValueError("Cannot provide both data and json_data")

        if data:
            kwargs["data"] = data

        elif json_data:
            kwargs["json"] = json_data

        return kwargs

    @clean_asana_response
    async def get(
        self,
        endpoint: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        api_version: str | None = None,
    ) -> dict:
        default_headers = {
            "Authorization": f"Bearer {self.auth_token}",
            "Accept": "application/json",
        }
        headers = {**default_headers, **(headers or {})}

        kwargs = {
            "url": self._build_url(endpoint, api_version),
            "headers": headers,
        }

        if params:
            kwargs["params"] = params

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.get(**kwargs)  # type: ignore[arg-type]
            self._raise_for_status(response)
        return cast(dict, response.json())

    @clean_asana_response
    async def post(
        self,
        endpoint: str,
        data: Optional[dict] = None,
        json_data: Optional[dict] = None,
        files: Optional[dict] = None,
        headers: Optional[dict] = None,
        api_version: str | None = None,
    ) -> dict:
        default_headers = {
            "Authorization": f"Bearer {self.auth_token}",
            "Accept": "application/json",
        }

        if files is None and json_data is not None:
            default_headers["Content-Type"] = "application/json"

        headers = {**default_headers, **(headers or {})}

        kwargs = {
            "url": self._build_url(endpoint, api_version),
            "headers": headers,
        }

        if files is not None:
            kwargs["files"] = files
            if data is not None:
                kwargs["data"] = data
        else:
            kwargs = self._set_request_body(kwargs, data, json_data)

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.post(**kwargs)  # type: ignore[arg-type]
            self._raise_for_status(response)
        return cast(dict, response.json())

    @clean_asana_response
    async def put(
        self,
        endpoint: str,
        data: Optional[dict] = None,
        json_data: Optional[dict] = None,
        headers: Optional[dict] = None,
        api_version: str | None = None,
    ) -> dict:
        headers = headers or {}
        headers["Authorization"] = f"Bearer {self.auth_token}"
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"

        kwargs = {
            "url": self._build_url(endpoint, api_version),
            "headers": headers,
        }

        kwargs = self._set_request_body(kwargs, data, json_data)

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.put(**kwargs)  # type: ignore[arg-type]
            self._raise_for_status(response)
        return cast(dict, response.json())

    async def get_current_user(self) -> dict:
        response = await self.get("/users/me")
        return cast(dict, response["data"])



from typing import TypeVar

from arcade_core.schema import ToolCallError, ToolCallLog, ToolCallOutput
from arcade_core.utils import coerce_empty_list_to_none

T = TypeVar("T")


class ToolOutputFactory:
    """
    Singleton pattern for unified return method from tools.
    """

    def success(
        self,
        *,
        data: T | None = None,
        logs: list[ToolCallLog] | None = None,
    ) -> ToolCallOutput:
        value = getattr(data, "result", "") if data else ""
        logs = coerce_empty_list_to_none(logs)
        return ToolCallOutput(value=value, logs=logs)

    def fail(
        self,
        *,
        message: str,
        developer_message: str | None = None,
        traceback_info: str | None = None,
        logs: list[ToolCallLog] | None = None,
    ) -> ToolCallOutput:
        return ToolCallOutput(
            error=ToolCallError(
                message=message,
                developer_message=developer_message,
                can_retry=False,
                traceback_info=traceback_info,
            ),
            logs=coerce_empty_list_to_none(logs),
        )

    def fail_retry(
        self,
        *,
        message: str,
        developer_message: str | None = None,
        additional_prompt_content: str | None = None,
        retry_after_ms: int | None = None,
        traceback_info: str | None = None,
        logs: list[ToolCallLog] | None = None,
    ) -> ToolCallOutput:
        return ToolCallOutput(
            error=ToolCallError(
                message=message,
                developer_message=developer_message,
                can_retry=True,
                additional_prompt_content=additional_prompt_content,
                retry_after_ms=retry_after_ms,
            ),
            logs=coerce_empty_list_to_none(logs),
        )


output_factory = ToolOutputFactory()



from base64 import urlsafe_b64encode
from email.message import EmailMessage
from unittest.mock import MagicMock, patch

import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext
from arcade_tdk.errors import ToolExecutionError
from googleapiclient.errors import HttpError

from arcade_gmail.enums import GmailReplyToWhom
from arcade_gmail.tools import (
    delete_draft_email,
    get_thread,
    list_draft_emails,
    list_emails,
    list_emails_by_header,
    list_threads,
    reply_to_email,
    search_threads,
    send_draft_email,
    send_email,
    trash_email,
    update_draft_email,
    write_draft_email,
)
from arcade_gmail.utils import (
    build_reply_body,
    parse_draft_email,
    parse_multipart_email,
    parse_plain_text_email,
)


@pytest.fixture
def mock_context():
    mock_auth = ToolAuthorizationContext(token="fake-token")  # noqa: S106
    return ToolContext(authorization=mock_auth)


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_send_email(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Test happy path
    result = await send_email(
        context=mock_context,
        subject="Test Subject",
        body="Test Body",
        recipient="test@example.com",
    )

    assert isinstance(result, dict)
    assert "id" in result
    assert "thread_id" in result
    assert "subject" in result
    assert "body" in result

    # Test http error
    mock_service.users().messages().send().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid recipient"}}',
    )

    with pytest.raises(ToolExecutionError):
        await send_email(
            context=mock_context,
            subject="Test Subject",
            body="Test Body",
            recipient="invalid@example.com",
        )


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_write_draft_email(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Test happy path
    result = await write_draft_email(
        context=mock_context,
        subject="Test Draft Subject",
        body="Test Draft Body",
        recipient="draft@example.com",
    )

    assert isinstance(result, dict)
    assert "id" in result
    assert "thread_id" in result
    assert "subject" in result
    assert "body" in result

    # Test http error
    mock_service.users().drafts().create().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await write_draft_email(
            context=mock_context,
            subject="Test Draft Subject",
            body="Test Draft Body",
            recipient="draft@example.com",
        )


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_update_draft_email(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Test happy path
    result = await update_draft_email(
        context=mock_context,
        draft_email_id="draft123",
        subject="Updated Subject",
        body="Updated Body",
        recipient="updated@example.com",
    )

    assert isinstance(result, dict)
    assert "id" in result
    assert "thread_id" in result
    assert "subject" in result
    assert "body" in result

    # Test http error
    mock_service.users().drafts().update().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Draft not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await update_draft_email(
            context=mock_context,
            draft_email_id="nonexistent_draft",
            subject="Updated Subject",
            body="Updated Body",
            recipient="updated@example.com",
        )


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_send_draft_email(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Test happy path
    result = await send_draft_email(context=mock_context, email_id="draft456")

    assert isinstance(result, dict)
    assert "id" in result
    assert "thread_id" in result
    assert "subject" in result
    assert "body" in result

    # Test http error
    mock_service.users().drafts().send().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Draft not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await send_draft_email(context=mock_context, email_id="nonexistent_draft")


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_delete_draft_email(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Test happy path
    result = await delete_draft_email(context=mock_context, draft_email_id="draft789")

    assert "Draft email with ID" in result
    assert "deleted successfully" in result

    # Test http error
    mock_service.users().drafts().delete().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Draft not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await delete_draft_email(context=mock_context, draft_email_id="nonexistent_draft")


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
@patch("arcade_gmail.tools.gmail.parse_draft_email")
async def test_get_draft_emails(mock_parse_draft_email, mock_build, mock_context):
    # Setup test data
    mock_drafts_list_response = {
        "drafts": [
            {
                "id": "r9999999999999999999",
                "message": {"id": "0000000000000000", "threadId": "0000000000000000"},
            }
        ],
        "resultSizeEstimate": 1,
    }
    mock_drafts_get_response = {
        "id": "r9999999999999999999",
        "message": {
            "id": "0000000000000000",
            "threadId": "0000000000000000",
            "labelIds": ["DRAFT"],
            "snippet": "Hello! This is a test. Best regards, John",
            "payload": {
                "partId": "",
                "mimeType": "text/plain",
                "filename": "",
                "headers": [
                    {"name": "to", "value": "test@arcade-ai.com"},
                    {"name": "subject", "value": "New Draft"},
                    {"name": "Date", "value": "Mon, 16 Sep 2024 13:02:10 -0400"},
                    {"name": "From", "value": "john-doe@arcade-ai.com"},
                ],
                "body": {
                    "size": 41,
                    "data": "SGVsbG8hIFRoaXMgaXMgYSB0ZXN0LgoKQmVzdCByZWdhcmRzLApCb2I=",
                },
            },
            "sizeEstimate": 453,
            "historyId": "7061",
            "internalDate": "1726506130000",
        },
    }

    # Setup mocking
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Mock the response from the Gmail list drafts API
    mock_service.users().drafts().list().execute.return_value = mock_drafts_list_response

    # Mock the response from the Gmail get drafts API
    mock_service.users().drafts().get().execute.return_value = mock_drafts_get_response

    # Mock the parse_draft_email function since parse_draft_email doesn't accept object of type MagicMock
    mock_parse_draft_email.return_value = parse_draft_email(mock_drafts_get_response)

    # Test happy path
    result = await list_draft_emails(context=mock_context, n_drafts=2)

    assert isinstance(result, dict)
    assert "emails" in result
    assert len(result["emails"]) == 1
    assert all("id" in draft and "subject" in draft for draft in result["emails"])

    # Test http error
    mock_service.users().drafts().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await list_draft_emails(context=mock_context, n_drafts=2)


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
@patch("arcade_gmail.tools.gmail.parse_plain_text_email")
async def test_search_emails_by_header(mock_parse_plain_text_email, mock_build, mock_context):
    # Setup test data
    mock_messages_list_response = {
        "messages": [
            {"id": "191fbc8ddce0f433", "threadId": "191fbc8ddce0f433"},
            {"id": "191fbc0ea11efa90", "threadId": "191fbc0ea11efa90"},
        ],
        "nextPageToken": "00755945214480102915",
        "resultSizeEstimate": 201,
    }
    mock_messages_get_response = {
        "id": "191f2cf4d24bf23d",
        "threadId": "191f2cf4d24bf23d",
        "labelIds": ["UNREAD", "IMPORTANT", "CATEGORY_UPDATES", "INBOX"],
        "snippet": "Hey User, Your personal access token (classic) &quot;ArcadeAI&quot; with admin:enterprise, admin:gpg_key, admin:org, admin:org_hook, admin:public_key, admin:repo_hook, admin:ssh_signing_key,",
        "payload": {
            "partId": "",
            "mimeType": "text/plain",
            "filename": "",
            "headers": [
                {"name": "Delivered-To", "value": "example@arcade-ai.com"},
                {"name": "Date", "value": "Sat, 14 Sep 2024 16:12:37 -0700"},
                {"name": "From", "value": "GitHub \u003cnoreply@github.com\u003e"},
                {"name": "To", "value": "example@arcade-ai.com"},
                {
                    "name": "Subject",
                    "value": "[GitHub] Your personal access token (classic) has expired",
                },
            ],
            "body": {
                "size": 605,
                "data": "SGV5IEBFcmljR3VzdGluLA0KDQpZb3VyIHBlcnNvbmFsIGFjY2VzcyB0b2tlbiAoY2xhc3NpYykgIkFyY2FkZUFJIiB3aXRoIGFkbWluOmVudGVycHJpc2UsIGFkbWluOmdwZ19rZXksIGFkbWluOm9yZywgYWRtaW46b3JnX2hvb2ssIGFkbWluOnB1YmxpY19rZXksIGFkbWluOnJlcG9faG9vaywgYWRtaW46c3NoX3NpZ25pbmdfa2V5LCBhdWRpdF9sb2csIGNvZGVzcGFjZSwgY29waWxvdCwgZGVsZXRlOnBhY2thZ2VzLCBkZWxldGVfcmVwbywgZ2lzdCwgbm90aWZpY2F0aW9ucywgcHJvamVjdCwgcmVwbywgdXNlciwgd29ya2Zsb3csIHdyaXRlOmRpc2N1c3Npb24sIGFuZCB3cml0ZTpwYWNrYWdlcyBzY29wZXMgaGFzIGV4cGlyZWQuDQoNCklmIHRoaXMgdG9rZW4gaXMgc3RpbGwgbmVlZGVkLCB2aXNpdCBodHRwczovL2dpdGh1Yi5jb20vc2V0dGluZ3MvdG9rZW5zLzE3MTM2OTg2MTMvcmVnZW5lcmF0ZSB0byBnZW5lcmF0ZSBhbiBlcXVpdmFsZW50Lg0KDQpJZiB5b3UgcnVuIGludG8gcHJvYmxlbXMsIHBsZWFzZSBjb250YWN0IHN1cHBvcnQgYnkgdmlzaXRpbmcgaHR0cHM6Ly9naXRodWIuY29tL2NvbnRhY3QNCg0KVGhhbmtzLA0KVGhlIEdpdEh1YiBUZWFtDQo=",
            },
        },
        "sizeEstimate": 4512,
        "historyId": "5508",
        "internalDate": "1726355557000",
    }

    # Setup mocking
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Mock the response from the Gmail list messages API
    mock_service.users().messages().list().execute.return_value = mock_messages_list_response

    # Mock the response from the Gmail get messages API
    mock_service.users().messages().get().execute.return_value = mock_messages_get_response

    # Mock the parse_plain_text_email function since parse_plain_text_email doesn't accept object of type MagicMock
    mock_parse_plain_text_email.return_value = parse_plain_text_email(mock_messages_get_response)

    # Test happy path
    result = await list_emails_by_header(
        context=mock_context, sender="noreply@github.com", max_results=2
    )

    assert isinstance(result, dict)
    assert "emails" in result
    assert len(result["emails"]) == 2
    assert all("id" in email and "subject" in email for email in result["emails"])

    # Test http error
    mock_service.users().messages().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await list_emails_by_header(
            context=mock_context, sender="noreply@github.com", max_results=2
        )


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
@patch("arcade_gmail.tools.gmail.parse_plain_text_email")
async def test_get_emails(mock_parse_plain_text_email, mock_build, mock_context):
    # Setup test data
    mock_messages_list_response = {
        "messages": [
            {"id": "191fbc8ddce0f433", "threadId": "191fbc8ddce0f433"},
        ],
        "nextPageToken": "00755945214480102915",
        "resultSizeEstimate": 1,
    }
    mock_messages_get_response = {
        "id": "191f2cf4d24bf23d",
        "threadId": "191f2cf4d24bf23d",
        "labelIds": ["UNREAD", "IMPORTANT", "CATEGORY_UPDATES", "INBOX"],
        "snippet": "Hey User, Your personal access token (classic) &quot;ArcadeAI&quot; with admin:enterprise, admin:gpg_key, admin:org, admin:org_hook, admin:public_key, admin:repo_hook, admin:ssh_signing_key,",
        "payload": {
            "partId": "",
            "mimeType": "text/plain",
            "filename": "",
            "headers": [
                {"name": "Delivered-To", "value": "example@arcade-ai.com"},
                {"name": "Date", "value": "Sat, 14 Sep 2024 16:12:37 -0700"},
                {"name": "From", "value": "GitHub \u003cnoreply@github.com\u003e"},
                {"name": "To", "value": "example@arcade-ai.com"},
                {
                    "name": "Subject",
                    "value": "[GitHub] Your personal access token (classic) has expired",
                },
            ],
            "body": {
                "size": 605,
                "data": "SGV5IEBFcmljR3VzdGluLA0KDQpZb3VyIHBlcnNvbmFsIGFjY2VzcyB0b2tlbiAoY2xhc3NpYykgIkFyY2FkZUFJIiB3aXRoIGFkbWluOmVudGVycHJpc2UsIGFkbWluOmdwZ19rZXksIGFkbWluOm9yZywgYWRtaW46b3JnX2hvb2ssIGFkbWluOnB1YmxpY19rZXksIGFkbWluOnJlcG9faG9vaywgYWRtaW46c3NoX3NpZ25pbmdfa2V5LCBhdWRpdF9sb2csIGNvZGVzcGFjZSwgY29waWxvdCwgZGVsZXRlOnBhY2thZ2VzLCBkZWxldGVfcmVwbywgZ2lzdCwgbm90aWZpY2F0aW9ucywgcHJvamVjdCwgcmVwbywgdXNlciwgd29ya2Zsb3csIHdyaXRlOmRpc2N1c3Npb24sIGFuZCB3cml0ZTpwYWNrYWdlcyBzY29wZXMgaGFzIGV4cGlyZWQuDQoNCklmIHRoaXMgdG9rZW4gaXMgc3RpbGwgbmVlZGVkLCB2aXNpdCBodHRwczovL2dpdGh1Yi5jb20vc2V0dGluZ3MvdG9rZW5zLzE3MTM2OTg2MTMvcmVnZW5lcmF0ZSB0byBnZW5lcmF0ZSBhbiBlcXVpdmFsZW50Lg0KDQpJZiB5b3UgcnVuIGludG8gcHJvYmxlbXMsIHBsZWFzZSBjb250YWN0IHN1cHBvcnQgYnkgdmlzaXRpbmcgaHR0cHM6Ly9naXRodWIuY29tL2NvbnRhY3QNCg0KVGhhbmtzLA0KVGhlIEdpdEh1YiBUZWFtDQo=",
            },
        },
        "sizeEstimate": 4512,
        "historyId": "5508",
        "internalDate": "1726355557000",
    }

    # Setup mocking
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Mock the response from the Gmail list messages API
    mock_service.users().messages().list().execute.return_value = mock_messages_list_response

    # Mock the Gmail get messages API
    mock_service.users().messages().get().execute.return_value = mock_messages_get_response

    # Mock the parse_plain_text_email function since parse_plain_text_email doesn't accept object of type MagicMock
    mock_parse_plain_text_email.return_value = parse_plain_text_email(mock_messages_get_response)

    # Test happy path
    result = await list_emails(context=mock_context, n_emails=1)

    assert isinstance(result, dict)
    assert "emails" in result
    assert len(result["emails"]) == 1
    assert "id" in result["emails"][0]
    assert "subject" in result["emails"][0]
    assert "date" in result["emails"][0]
    assert "body" in result["emails"][0]

    # Test http error
    mock_service.users().messages().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await list_emails(context=mock_context, n_emails=1)


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_trash_email(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Test happy path
    email_id = "123456"
    result = await trash_email(context=mock_context, email_id=email_id)

    assert isinstance(result, dict)
    assert "id" in result
    assert "thread_id" in result
    assert "subject" in result
    assert "body" in result

    # Test http error
    mock_service.users().messages().trash().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Email not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await trash_email(context=mock_context, email_id="nonexistent_email")


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_search_threads(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Setup mock response data
    mock_threads_list_response = {
        "threads": [
            {
                "id": "thread1",
                "snippet": "Thread snippet 1",
            },
            {
                "id": "thread2",
                "snippet": "Thread snippet 2",
            },
        ],
        "nextPageToken": "next_token_123",
        "resultSizeEstimate": 2,
    }

    # Mock the Gmail API threads().list() method
    mock_service.users().threads().list().execute.return_value = mock_threads_list_response

    # Test happy path
    result = await search_threads(
        context=mock_context,
        sender="test@example.com",
        max_results=2,
    )

    assert isinstance(result, dict)
    assert "threads" in result
    assert len(result["threads"]) == 2
    assert result["threads"][0]["id"] == "thread1"
    assert "next_page_token" in result

    # Test error handling
    mock_service.users().threads().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await search_threads(
            context=mock_context,
            sender="test@example.com",
            max_results=2,
        )


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_list_threads(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Setup mock response data
    mock_threads_list_response = {
        "threads": [
            {
                "id": "thread1",
                "snippet": "Thread snippet 1",
            },
            {
                "id": "thread2",
                "snippet": "Thread snippet 2",
            },
        ],
        "nextPageToken": "next_token_123",
        "resultSizeEstimate": 2,
    }

    # Mock the Gmail API threads().list() method
    mock_service.users().threads().list().execute.return_value = mock_threads_list_response

    # Test happy path
    result = await list_threads(
        context=mock_context,
        max_results=2,
    )

    assert isinstance(result, dict)
    assert "threads" in result
    assert len(result["threads"]) == 2
    assert result["threads"][0]["id"] == "thread1"
    assert "next_page_token" in result

    # Test error handling
    mock_service.users().threads().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await list_threads(
            context=mock_context,
            max_results=2,
        )


@pytest.mark.asyncio
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_get_thread(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Setup mock response data
    mock_thread_get_response = {
        "id": "thread1",
        "messages": [
            {
                "id": "message1",
                "snippet": "Message snippet 1",
            },
            {
                "id": "message2",
                "snippet": "Message snippet 2",
            },
        ],
    }

    # Mock the Gmail API threads().get() method
    mock_service.users().threads().get().execute.return_value = mock_thread_get_response

    # Test happy path
    result = await get_thread(
        context=mock_context,
        thread_id="thread1",
    )

    assert isinstance(result, dict)
    assert "id" in result
    assert result["id"] == "thread1"
    assert "messages" in result
    assert len(result["messages"]) == 2
    assert result["messages"][0]["id"] == "message1"

    # Test error handling
    mock_service.users().threads().get().execute.side_effect = HttpError(
        resp=MagicMock(status=404),
        content=b'{"error": {"message": "Thread not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await get_thread(
            context=mock_context,
            thread_id="invalid_thread",
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "reply_to_whom, expected_to, expected_cc",
    [
        (
            GmailReplyToWhom.EVERY_RECIPIENT,
            "sender@example.com, to1@example.com, to2@example.com",
            "cc1@example.com, cc2@example.com",
        ),
        (
            GmailReplyToWhom.ONLY_THE_SENDER,
            "sender@example.com",
            "",
        ),
    ],
)
@patch("arcade_gmail.tools.gmail._build_gmail_service")
async def test_reply_to_email(mock_build, reply_to_whom, expected_to, expected_cc, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    original_message = {
        "id": "id123456",
        "threadId": "thread123456",
        "payload": {
            "headers": [
                {"name": "Message-ID", "value": "id123456"},
                {"name": "Subject", "value": "test"},
                {"name": "From", "value": "sender@example.com"},
                {"name": "To", "value": "to1@example.com, to2@example.com, test@example.com"},
                {"name": "Cc", "value": "cc1@example.com, cc2@example.com"},
                {"name": "References", "value": "thread123456"},
            ],
        },
    }

    mock_service.users().getProfile().execute.return_value = {"emailAddress": "test@example.com"}
    mock_service.users().messages().get().execute.return_value = original_message

    result = await reply_to_email(
        context=mock_context,
        body="test",
        reply_to_message_id="id123456",
        reply_to_whom=reply_to_whom,
    )

    assert isinstance(result, dict)
    assert "url" in result

    replying_to = parse_multipart_email(original_message)
    expected_body = build_reply_body("test", replying_to)

    expected_message = EmailMessage()
    expected_message.set_content(expected_body)
    expected_message["To"] = expected_to
    expected_message["Subject"] = "Re: test"
    if expected_cc:
        expected_message["Cc"] = expected_cc
    expected_message["In-Reply-To"] = "id123456"
    expected_message["References"] = "id123456, thread123456"

    mock_service.users().messages().send.assert_called_once_with(
        userId="me",
        body={
            "raw": urlsafe_b64encode(expected_message.as_bytes()).decode(),
            "threadId": "thread123456",
        },
    )


def test_parse_multipart_email_full():
    """
    Test parsing a multipart email with both plain text and HTML bodies.
    """
    email_data = {
        "id": "email123",
        "threadId": "thread123",
        "labelIds": ["INBOX", "UNREAD"],
        "historyId": "history123",
        "snippet": "This is a test email.",
        "payload": {
            "headers": [
                {"name": "To", "value": "recipient@example.com"},
                {"name": "From", "value": "sender@example.com"},
                {"name": "Subject", "value": "Test Email"},
                {"name": "Date", "value": "Mon, 1 Jan 2024 10:00:00 -0000"},
            ],
            "body": {"size": 100, "data": "VGhpcyBpcyBhIHRlc3QgZW1haWwu"},
        },
    }

    with (
        patch("arcade_gmail.utils._get_email_plain_text_body") as mock_plain,
        patch("arcade_gmail.utils._get_email_html_body") as mock_html,
        patch("arcade_gmail.utils._clean_email_body") as mock_clean,
    ):
        # Mock the helper functions
        mock_plain.return_value = "This is a test email."
        mock_html.return_value = "<p>This is a test email.</p>"
        mock_clean.return_value = "This is a test email."

        result = parse_multipart_email(email_data)

        assert result["id"] == "email123"
        assert result["thread_id"] == "thread123"
        assert result["label_ids"] == ["INBOX", "UNREAD"]
        assert result["snippet"] == "This is a test email."
        assert result["to"] == "recipient@example.com"
        assert result["from"] == "sender@example.com"
        assert result["subject"] == "Test Email"
        assert result["date"] == "Mon, 1 Jan 2024 10:00:00 -0000"
        assert result["plain_text_body"] == "This is a test email."
        assert result["html_body"] == "<p>This is a test email.</p>"


def test_parse_multipart_email_plain_only():
    """
    Test parsing an email with only a plain text body.
    """
    email_data = {
        "id": "email456",
        "threadId": "thread456",
        "labelIds": ["INBOX"],
        "historyId": "history456",
        "snippet": "Plain text only email.",
        "payload": {
            "headers": [
                {"name": "To", "value": "recipient2@example.com"},
                {"name": "From", "value": "sender2@example.com"},
                {"name": "Subject", "value": "Plain Text Email"},
                {"name": "Date", "value": "Tue, 2 Feb 2024 11:00:00 -0000"},
            ],
            "body": {"size": 150, "data": "UGxhaW4gdGV4dCBvbmx5IGVtYWlsLg=="},
        },
    }

    with (
        patch("arcade_gmail.utils._get_email_plain_text_body") as mock_plain,
        patch("arcade_gmail.utils._get_email_html_body") as mock_html,
        patch("arcade_gmail.utils._clean_email_body") as mock_clean,
    ):
        # Mock the helper functions
        mock_plain.return_value = "Plain text only email."
        mock_html.return_value = None
        mock_clean.return_value = "Plain text only email."

        result = parse_multipart_email(email_data)

        assert result["id"] == "email456"
        assert result["thread_id"] == "thread456"
        assert result["label_ids"] == ["INBOX"]
        assert result["snippet"] == "Plain text only email."
        assert result["to"] == "recipient2@example.com"
        assert result["from"] == "sender2@example.com"
        assert result["subject"] == "Plain Text Email"
        assert result["date"] == "Tue, 2 Feb 2024 11:00:00 -0000"
        assert result["plain_text_body"] == "Plain text only email."
        assert result["html_body"] == ""


def test_parse_multipart_email_html_only():
    """
    Test parsing an email with only an HTML body.
    """
    email_data = {
        "id": "email789",
        "threadId": "thread789",
        "labelIds": ["SENT"],
        "historyId": "history789",
        "snippet": "HTML only email.",
        "payload": {
            "headers": [
                {"name": "To", "value": "recipient3@example.com"},
                {"name": "From", "value": "sender3@example.com"},
                {"name": "Subject", "value": "HTML Email"},
                {"name": "Date", "value": "Wed, 3 Mar 2024 12:00:00 -0000"},
            ],
            "body": {"size": 200, "data": "PGh0bWw+VGhpcyBpcyBIVE1MIGVtYWlsLjwvaHRtbD4="},
        },
    }

    with (
        patch("arcade_gmail.utils._get_email_plain_text_body") as mock_plain,
        patch("arcade_gmail.utils._get_email_html_body") as mock_html,
        patch("arcade_gmail.utils._clean_email_body") as mock_clean,
    ):
        # Mock the helper functions
        mock_plain.return_value = None
        mock_html.return_value = "<html>This is HTML email.</html>"
        mock_clean.return_value = "This is HTML email."

        result = parse_multipart_email(email_data)

        assert result["id"] == "email789"
        assert result["thread_id"] == "thread789"
        assert result["label_ids"] == ["SENT"]
        assert result["snippet"] == "HTML only email."
        assert result["to"] == "recipient3@example.com"
        assert result["from"] == "sender3@example.com"
        assert result["subject"] == "HTML Email"
        assert result["date"] == "Wed, 3 Mar 2024 12:00:00 -0000"
        assert result["plain_text_body"] == "This is HTML email."
        assert result["html_body"] == "<html>This is HTML email.</html>"


def test_parse_multipart_email_missing_payload():
    """
    Test parsing an email with missing payload.
    """
    email_data = {
        "id": "email000",
        "threadId": "thread000",
        "labelIds": ["INBOX"],
        "historyId": "history000",
        "snippet": "Missing payload email.",
        # 'payload' key is missing
    }

    result = parse_multipart_email(email_data)

    # Since payload is missing, headers and bodies should be default or empty
    assert result["id"] == "email000"
    assert result["thread_id"] == "thread000"
    assert result["label_ids"] == ["INBOX"]
    assert result["snippet"] == "Missing payload email."
    assert result["to"] == ""
    assert result["from"] == ""
    assert result["subject"] == ""
    assert result["date"] == ""
    assert result["plain_text_body"] == ""
    assert result["html_body"] == ""


def test_parse_multipart_email_missing_headers():
    """
    Test parsing an email with missing headers in the payload.
    """
    email_data = {
        "id": "email111",
        "threadId": "thread111",
        "labelIds": ["INBOX"],
        "historyId": "history111",
        "snippet": "Missing headers email.",
        "payload": {
            # 'headers' key is missing
            "body": {"size": 100, "data": "VGltZWw="}
        },
    }

    with (
        patch("arcade_gmail.utils._get_email_plain_text_body") as mock_plain,
        patch("arcade_gmail.utils._get_email_html_body") as mock_html,
        patch("arcade_gmail.utils._clean_email_body") as mock_clean,
    ):
        # Mock the helper functions
        mock_plain.return_value = "Timeel"
        mock_html.return_value = "<p>Timeel</p>"
        mock_clean.return_value = "Timeel"

        result = parse_multipart_email(email_data)

    assert result["id"] == "email111"
    assert result["thread_id"] == "thread111"
    assert result["label_ids"] == ["INBOX"]
    assert result["snippet"] == "Missing headers email."
    assert result["to"] == ""
    assert result["from"] == ""
    assert result["subject"] == ""
    assert result["date"] == ""
    assert result["plain_text_body"] == "Timeel"
    assert result["html_body"] == "<p>Timeel</p>"


def test_parse_multipart_email_missing_fields():
    """
    Test parsing an email with some missing fields in headers.
    """
    email_data = {
        "id": "email222",
        "threadId": "thread222",
        "labelIds": ["INBOX"],
        "historyId": "history222",
        "snippet": "Missing some headers.",
        "payload": {
            "headers": [
                {"name": "From", "value": "sender4@example.com"},
                {"name": "Subject", "value": "Partial Headers"},
                # 'To' and 'Date' headers are missing
            ],
            "body": {"size": 100, "data": "TWlzc2luZyBzb21lIGhlYWRlcnMu"},
        },
    }

    with (
        patch("arcade_gmail.utils._get_email_plain_text_body") as mock_plain,
        patch("arcade_gmail.utils._get_email_html_body") as mock_html,
        patch("arcade_gmail.utils._clean_email_body") as mock_clean,
    ):
        # Mock the helper functions
        mock_plain.return_value = "Missing some headers."
        mock_html.return_value = None
        mock_clean.return_value = "Missing some headers."

        result = parse_multipart_email(email_data)

    assert result["id"] == "email222"
    assert result["thread_id"] == "thread222"
    assert result["label_ids"] == ["INBOX"]
    assert result["snippet"] == "Missing some headers."
    assert result["to"] == ""
    assert result["from"] == "sender4@example.com"
    assert result["subject"] == "Partial Headers"
    assert result["date"] == ""
    assert result["plain_text_body"] == "Missing some headers."
    assert result["html_body"] == ""


def test_parse_multipart_email_empty():
    """
    Test parsing an empty email data.
    """
    email_data = {}

    result = parse_multipart_email(email_data)

    assert result["id"] == ""
    assert result["thread_id"] == ""
    assert result["label_ids"] == []
    assert result["snippet"] == ""
    assert result["to"] == ""
    assert result["from"] == ""
    assert result["subject"] == ""
    assert result["date"] == ""
    assert result["plain_text_body"] == ""
    assert result["html_body"] == ""


def test_parse_multipart_email_invalid_payload_structure():
    """
    Test parsing an email with an invalid payload structure.
    """
    email_data = {
        "id": "email333",
        "threadId": "thread333",
        "labelIds": ["INBOX"],
        "historyId": "history333",
        "snippet": "Invalid payload structure.",
        "payload": {
            "headers": "This should be a list, not a string",
            "body": {"size": 100, "data": "SW52YWxpZCBwYXlsb2Fk"},
        },
    }

    with pytest.raises(TypeError):
        parse_multipart_email(email_data)



import decimal
import math
from decimal import Decimal
from typing import Annotated

from arcade_tdk import tool

decimal.getcontext().prec = 100


@tool
def ceil(
    a: Annotated[str, "The number to round up as a string"],
) -> Annotated[str, "The smallest integer greater than or equal to the number as a string"]:
    """
    Return the ceiling of a number
    """
    # Use Decimal for arbitrary precision
    return str(math.ceil(Decimal(a)))


@tool
def floor(
    a: Annotated[str, "The number to round down as a string"],
) -> Annotated[str, "The largest integer less than or equal to the number as a string"]:
    """
    Return the floor of a number
    """
    # Use Decimal for arbitrary precision
    return str(math.floor(Decimal(a)))


@tool
def round_num(
    value: Annotated[str, "The number to round as a string"],
    ndigits: Annotated[str, "The number of digits after the decimal point as a string"],
) -> Annotated[str, "The number rounded to the specified number of digits as a string"]:
    """
    Round a number to a specified number of positive digits
    """
    ndigits_int = int(ndigits)
    if ndigits_int >= 0:
        # Use Decimal for arbitrary precision
        return str(round(Decimal(value), int(ndigits_int)))
    # cast value from str -> float -> int here because rounding with negative
    # decimals is only useful for weird math
    return str(round(int(float(value)), int(ndigits_int)))



import contextlib
import re
from datetime import datetime
from typing import Any, cast
from urllib.parse import parse_qs, urlparse
from zoneinfo import ZoneInfo

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError
from serpapi import Client as SerpClient

from arcade_search.constants import (
    DEFAULT_GOOGLE_COUNTRY,
    DEFAULT_GOOGLE_LANGUAGE,
    DEFAULT_GOOGLE_MAPS_COUNTRY,
    DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
    DEFAULT_GOOGLE_MAPS_LANGUAGE,
    DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
    YOUTUBE_MAX_DESCRIPTION_LENGTH,
)
from arcade_search.enums import GoogleMapsDistanceUnit, GoogleMapsTravelMode
from arcade_search.exceptions import CountryNotFoundError, LanguageNotFoundError
from arcade_search.google_data import COUNTRY_CODES, LANGUAGE_CODES


# ------------------------------------------------------------------------------------------------
# General SerpAPI utils
# ------------------------------------------------------------------------------------------------
def prepare_params(engine: str, **kwargs: Any) -> dict[str, Any]:
    """
    Prepares a parameters dictionary for the SerpAPI call.

    Parameters:
        engine: The engine name (e.g., "google", "google_finance").
        kwargs: Any additional parameters to include.

    Returns:
        A dictionary containing the base parameters plus any extras,
        excluding any parameters whose value is None.
    """
    params = {"engine": engine}
    params.update({k: v for k, v in kwargs.items() if v is not None})
    return params


def call_serpapi(context: ToolContext, params: dict) -> dict:
    """
    Execute a search query using the SerpAPI client and return the results as a dictionary.

    Args:
        context: The tool context containing required secrets.
        params: A dictionary of parameters for the SerpAPI search.

    Returns:
        The search results as a dictionary.
    """
    api_key = context.get_secret("SERP_API_KEY")
    client = SerpClient(api_key=api_key)
    try:
        search = client.search(params)
        return cast(dict[str, Any], search.as_dict())
    except Exception as e:
        # SerpAPI error messages sometimes contain the API key, so we need to sanitize it
        sanitized_e = re.sub(r"(api_key=)[^ &]+", r"\1***", str(e))
        raise ToolExecutionError(
            message="Failed to fetch search results",
            developer_message=sanitized_e,
        )


# ------------------------------------------------------------------------------------------------
# Google general utils
# ------------------------------------------------------------------------------------------------
def default_language_code(default_service_language_code: str | None = None) -> str | None:
    if isinstance(default_service_language_code, str):
        return default_service_language_code.lower()
    elif isinstance(DEFAULT_GOOGLE_LANGUAGE, str):
        return DEFAULT_GOOGLE_LANGUAGE.lower()
    return None


def default_country_code(default_service_country_code: str | None = None) -> str | None:
    if isinstance(default_service_country_code, str):
        return default_service_country_code.lower()
    elif isinstance(DEFAULT_GOOGLE_COUNTRY, str):
        return DEFAULT_GOOGLE_COUNTRY.lower()
    return None


def resolve_language_code(
    language_code: str | None = None,
    default_service_language_code: str | None = None,
) -> str | None:
    language_code = language_code or default_language_code(default_service_language_code)

    if isinstance(language_code, str):
        language_code = language_code.lower()
        if language_code not in LANGUAGE_CODES:
            raise LanguageNotFoundError(language_code)

    return language_code


def resolve_country_code(
    country_code: str | None = None,
    default_service_country_code: str | None = None,
) -> str | None:
    country_code = country_code or default_country_code(default_service_country_code)

    if isinstance(country_code, str):
        country_code = country_code.lower()
        if country_code not in COUNTRY_CODES:
            raise CountryNotFoundError(country_code)

    return country_code


# ------------------------------------------------------------------------------------------------
# Google Maps utils
# ------------------------------------------------------------------------------------------------
def get_google_maps_directions(
    context: ToolContext,
    origin_address: str | None = None,
    destination_address: str | None = None,
    origin_latitude: str | None = None,
    origin_longitude: str | None = None,
    destination_latitude: str | None = None,
    destination_longitude: str | None = None,
    language: str | None = DEFAULT_GOOGLE_MAPS_LANGUAGE,
    country: str | None = DEFAULT_GOOGLE_MAPS_COUNTRY,
    distance_unit: GoogleMapsDistanceUnit = DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
    travel_mode: GoogleMapsTravelMode = DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
) -> list[dict[str, Any]]:
    """Get directions from Google Maps.

    Provide either all(origin_address, destination_address) or
    all(origin_latitude, origin_longitude, destination_latitude, destination_longitude).

    Args:
        context: Tool context containing required Serp API Key secret.
        origin_address: Origin address.
        destination_address: Destination address.
        origin_latitude: Origin latitude.
        origin_longitude: Origin longitude.
        destination_latitude: Destination latitude.
        destination_longitude: Destination longitude.
        language: Language to use in the Google Maps search. Defaults to 'en' (English).
        country: 2-letter country code to use in the Google Maps search. Defaults to None
            (no country is specified).
        distance_unit: Distance unit to use in the Google Maps search. Defaults to 'km'
            (kilometers).
        travel_mode: Travel mode to use in the Google Maps search. Defaults to 'best'
            (best mode).

    Returns:
        The directions from Google Maps.
    """
    if isinstance(language, str):
        language = language.lower()

    if language not in LANGUAGE_CODES:
        raise LanguageNotFoundError(language)

    params = prepare_params(
        engine="google_maps_directions",
        hl=language,
        distance_unit=distance_unit.to_api_value(),
        travel_mode=travel_mode.to_api_value(),
    )

    if any([
        origin_latitude,
        origin_longitude,
        destination_latitude,
        destination_longitude,
    ]) and any([origin_address, destination_address]):
        raise ValueError("Either coordinates or addresses must be provided, not both")

    elif all([origin_latitude, origin_longitude, destination_latitude, destination_longitude]):
        params["start_coords"] = f"{origin_latitude},{origin_longitude}"
        params["end_coords"] = f"{destination_latitude},{destination_longitude}"

    elif all([origin_address, destination_address]):
        params["start_addr"] = str(origin_address)
        params["end_addr"] = str(destination_address)

    else:
        raise ValueError("Either coordinates or addresses must be provided")

    if country:
        country = country.lower()
        if country not in COUNTRY_CODES:
            raise CountryNotFoundError(country)
        params["gl"] = country

    results = call_serpapi(context, params)

    directions = cast(list[dict[str, Any]], results.get("directions", []))

    for direction in directions:
        clean_google_maps_direction(direction)

        if "arrive_around" in direction:
            direction["arrive_around"] = enrich_google_maps_arrive_around(
                direction["arrive_around"]
            )

    return directions


def clean_google_maps_direction(direction: dict[str, Any]) -> None:
    for trip in direction.get("trips", []):
        with contextlib.suppress(KeyError):
            del trip["start_stop"]["data_id"]
            del trip["end_stop"]["data_id"]

        for detail in trip.get("details", []):
            with contextlib.suppress(KeyError):
                del detail["geo_photo"]
                del detail["gps_coordinates"]

        for stop in trip.get("stops", []):
            with contextlib.suppress(KeyError):
                del stop["data_id"]


def enrich_google_maps_arrive_around(timestamp: int | None) -> dict[str, Any]:
    if not timestamp:
        return {}

    dt = datetime.fromtimestamp(timestamp, tz=ZoneInfo("UTC")).isoformat()
    return {"datetime": dt, "timestamp": timestamp}


# ------------------------------------------------------------------------------------------------
# Google Flights utils
# ------------------------------------------------------------------------------------------------
def parse_flight_results(results: dict[str, Any]) -> dict[str, Any]:
    """Parse the flight results from the Google Flights API

    Note: Best flights is not always returned from the API.
    """
    flight_data = {}
    flights = []

    if "best_flights" in results:
        flights.extend(results["best_flights"])
    if "other_flights" in results:
        flights.extend(results["other_flights"])
    if "price_insights" in results:
        flight_data["price_insights"] = results["price_insights"]

    flight_data["flights"] = flights

    return flight_data


# ------------------------------------------------------------------------------------------------
# Google News utils
# ------------------------------------------------------------------------------------------------
def extract_news_results(results: dict[str, Any], limit: int | None = None) -> list[dict[str, Any]]:
    news_results = []
    for result in results.get("news_results", []):
        news_results.append({
            "title": result.get("title"),
            "snippet": result.get("snippet"),
            "link": result.get("link"),
            "date": result.get("date"),
            "source": result.get("source", {}).get("name"),
        })

    if limit:
        return news_results[:limit]
    return news_results


# ------------------------------------------------------------------------------------------------
# Google Shopping utils
# ------------------------------------------------------------------------------------------------
def extract_shopping_results(results: list[dict[str, Any]]) -> list[dict[str, Any]]:
    return [
        {
            "title": result.get("title"),
            "direct_link": result.get("link"),
            "google_link": result.get("product_link"),
            "source": result.get("source"),
            "price": result.get("price"),
            "product_rating": result.get("rating"),
            "product_reviews": result.get("reviews"),
            "store_rating": result.get("store_rating"),
            "store_reviews": result.get("store_reviews"),
            "delivery": result.get("delivery"),
        }
        for result in results
    ]


# ------------------------------------------------------------------------------------------------
# Walmart utils
# ------------------------------------------------------------------------------------------------
def extract_walmart_results(results: list[dict[str, Any]]) -> list[dict[str, Any]]:
    return [
        {
            "item_id": result.get("us_item_id"),
            "title": result.get("title"),
            "description": result.get("description"),
            "rating": result.get("rating"),
            "reviews_count": result.get("reviews"),
            "seller": {
                "id": result.get("seller_id"),
                "name": result.get("seller_name"),
            },
            "price": {
                "value": result.get("primary_offer", {}).get("offer_price"),
                "currency": result.get("primary_offer", {}).get("offer_currency"),
            },
            "link": result.get("product_page_url"),
        }
        for result in results
    ]


def get_walmart_last_page_integer(results: dict[str, Any]) -> int:
    try:
        return int(list(results["pagination"]["other_pages"].keys())[-1])
    except (KeyError, IndexError, ValueError):
        return 1


def extract_walmart_product_details(product: dict[str, Any]) -> dict[str, Any]:
    return {
        "item_id": product.get("us_item_id"),
        "product_type": product.get("product_type"),
        "title": product.get("title"),
        "description_html": product.get("short_description_html"),
        "rating": product.get("rating"),
        "reviews_count": product.get("reviews"),
        "seller": {
            "id": product.get("seller_id"),
            "name": product.get("seller_name"),
        },
        "manufacturer_name": product.get("manufacturer"),
        "price": {
            "value": product.get("price_map", {}).get("price"),
            "currency": product.get("price_map", {}).get("currency"),
            "previous_price": product.get("price_map", {}).get("was_price", {}).get("price"),
        },
        "link": product.get("product_page_url"),
        "variant_options": extract_walmart_variant_options(product.get("variant_swatches", [])),
    }


def extract_walmart_variant_options(variant_swatches: list[dict[str, Any]]) -> list[dict[str, Any]]:
    variants = []

    for variant_swatch in variant_swatches:
        variant_name = variant_swatch.get("name")
        if not variant_name:
            continue

        options = []

        for selection in variant_swatch.get("available_selections", []):
            selection_name = selection.get("name")
            if selection_name and selection_name not in options:
                options.append(selection_name)

        variants.append({variant_name: options})

    return variants


# ------------------------------------------------------------------------------------------------
# YouTube utils
# ------------------------------------------------------------------------------------------------
def extract_video_id_from_link(link: str | None) -> str | None:
    if not isinstance(link, str):
        return None

    parsed_url = urlparse(link)
    query_params = parse_qs(parsed_url.query)
    return query_params.get("v", [""])[0]


def extract_video_description(
    video: dict[str, Any],
    max_description_length: int = YOUTUBE_MAX_DESCRIPTION_LENGTH,
) -> str | None:
    description = video.get("description", "")

    if isinstance(description, dict):
        description = description.get("content", "")

    if isinstance(description, str):
        too_long = len(description) > max_description_length
        if too_long:
            description = description[:max_description_length] + " [truncated]"

    if description is not None:
        description = str(description).strip()

    return cast(str | None, description)


def extract_video_results(
    results: dict[str, Any],
    max_description_length: int = YOUTUBE_MAX_DESCRIPTION_LENGTH,
) -> list[dict[str, Any]]:
    videos = []

    for video in results.get("video_results", []):
        videos.append({
            "id": extract_video_id_from_link(video.get("link")),
            "title": video.get("title"),
            "description": extract_video_description(video, max_description_length),
            "link": video.get("link"),
            "published_date": video.get("published_date"),
            "duration": video.get("duration"),
            "channel": {
                "name": video.get("channel", {}).get("name"),
                "link": video.get("channel", {}).get("link"),
            },
        })

    return videos


def extract_video_details(video: dict[str, Any]) -> dict[str, Any]:
    return {
        "id": extract_video_id_from_link(video.get("link")),
        "title": video.get("title"),
        "description": extract_video_description(video, YOUTUBE_MAX_DESCRIPTION_LENGTH),
        "published_date": video.get("published_date"),
        "channel": {
            "name": video.get("channel", {}).get("name"),
            "link": video.get("channel", {}).get("link"),
        },
        "like_count": video.get("extracted_likes"),
        "view_count": video.get("extracted_views"),
        "live": video.get("live", False),
    }



from arcade_google_sheets.tools.read import get_spreadsheet
from arcade_google_sheets.tools.write import create_spreadsheet, write_to_cell

__all__ = ["create_spreadsheet", "get_spreadsheet", "write_to_cell"]



import json
from typing import Annotated

from arcade_tdk import ToolContext, tool

from arcade_search.utils import call_serpapi, prepare_params


@tool(requires_secrets=["SERP_API_KEY"])
async def search_google(
    context: ToolContext,
    query: Annotated[str, "Search query"],
    n_results: Annotated[int, "Number of results to retrieve"] = 5,
) -> str:
    """Search Google using SerpAPI and return organic search results."""

    params = prepare_params("google", q=query)
    results = call_serpapi(context, params)
    organic_results = results.get("organic_results", [])

    return json.dumps(organic_results[:n_results])



import json
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import Optional
from zoneinfo import ZoneInfo

from pydantic import BaseModel, field_validator, model_validator


# ---------------------------------------------------------------------------- #
# Google Calendar Models and Enums
# ---------------------------------------------------------------------------- #
class DateRange(Enum):
    TODAY = "today"
    TOMORROW = "tomorrow"
    THIS_WEEK = "this_week"
    NEXT_WEEK = "next_week"
    THIS_MONTH = "this_month"
    NEXT_MONTH = "next_month"

    def to_datetime_range(
        self,
        start_time: time | None = None,
        end_time: time | None = None,
        time_zone: ZoneInfo | None = None,
        today: date | None = None,
    ) -> tuple[datetime, datetime]:
        """
        Convert a DateRange enum value to a tuple with two datetime objects representing the start
        and end of the date range.

        :param start_time: The start time of the date range. Defaults to the current time.
        :param end_time: The end time of the date range. Defaults to 23:59:59.
        :param time_zone: The time zone to use for the date range. Defaults to UTC.
        :param today: Today's date. Defaults to the current date provided by `datetime.now().date()`
        """
        start_time = start_time or datetime.now().time()
        end_time = end_time or time(23, 59, 59)
        today = today or datetime.now().date()

        if self == DateRange.TODAY:
            start_date, end_date = today, today
        elif self == DateRange.TOMORROW:
            start_date, end_date = today + timedelta(days=1), today + timedelta(days=1)
        elif self == DateRange.THIS_WEEK:
            start_date = today - timedelta(days=today.weekday())
            end_date = start_date + timedelta(days=6)
        elif self == DateRange.NEXT_WEEK:
            start_date = today + timedelta(days=7 - today.weekday())
            end_date = start_date + timedelta(days=6)
        elif self == DateRange.THIS_MONTH:
            start_date = today.replace(day=1)
            next_month = start_date + timedelta(days=31)
            end_date = next_month.replace(day=1) - timedelta(days=1)
        elif self == DateRange.NEXT_MONTH:
            start_date = (today.replace(day=1) + timedelta(days=31)).replace(day=1)
            next_month = start_date + timedelta(days=31)
            end_date = next_month.replace(day=1) - timedelta(days=1)
        else:
            raise ValueError(
                f"DateRange enum value: {self} is not supported for date range conversion"
            )

        start_time = start_time or time(0, 0, 0)
        end_time = end_time or time(23, 59, 59)

        start_datetime = datetime.combine(start_date, start_time)
        end_datetime = datetime.combine(end_date, end_time)

        if time_zone:
            start_datetime = start_datetime.replace(tzinfo=time_zone)
            end_datetime = end_datetime.replace(tzinfo=time_zone)

        return start_datetime, end_datetime


class Day(Enum):
    # TODO: THere are obvious limitations here. We should do better and support any date.
    YESTERDAY = "yesterday"
    TODAY = "today"
    TOMORROW = "tomorrow"
    THIS_SUNDAY = "this_sunday"
    THIS_MONDAY = "this_monday"
    THIS_TUESDAY = "this_tuesday"
    THIS_WEDNESDAY = "this_wednesday"
    THIS_THURSDAY = "this_thursday"
    THIS_FRIDAY = "this_friday"
    THIS_SATURDAY = "this_saturday"
    NEXT_SUNDAY = "next_sunday"
    NEXT_MONDAY = "next_monday"
    NEXT_TUESDAY = "next_tuesday"
    NEXT_WEDNESDAY = "next_wednesday"
    NEXT_THURSDAY = "next_thursday"
    NEXT_FRIDAY = "next_friday"
    NEXT_SATURDAY = "next_saturday"

    def to_date(self, time_zone_name: str) -> date:
        time_zone = ZoneInfo(time_zone_name)
        today = datetime.now(time_zone).date()
        weekday = today.weekday()

        if self == Day.YESTERDAY:
            return today - timedelta(days=1)
        elif self == Day.TODAY:
            return today
        elif self == Day.TOMORROW:
            return today + timedelta(days=1)

        day_offsets = {
            Day.THIS_SUNDAY: 6,
            Day.THIS_MONDAY: 0,
            Day.THIS_TUESDAY: 1,
            Day.THIS_WEDNESDAY: 2,
            Day.THIS_THURSDAY: 3,
            Day.THIS_FRIDAY: 4,
            Day.THIS_SATURDAY: 5,
        }

        if self in day_offsets:
            return today + timedelta(days=(day_offsets[self] - weekday) % 7)

        next_week_offsets = {
            Day.NEXT_SUNDAY: 6,
            Day.NEXT_MONDAY: 0,
            Day.NEXT_TUESDAY: 1,
            Day.NEXT_WEDNESDAY: 2,
            Day.NEXT_THURSDAY: 3,
            Day.NEXT_FRIDAY: 4,
            Day.NEXT_SATURDAY: 5,
        }

        if self in next_week_offsets:
            return today + timedelta(days=(next_week_offsets[self] - weekday + 7) % 7)

        raise ValueError(f"Invalid Day enum value: {self}")


class TimeSlot(Enum):
    _0000 = "00:00"
    _0015 = "00:15"
    _0030 = "00:30"
    _0045 = "00:45"
    _0100 = "01:00"
    _0115 = "01:15"
    _0130 = "01:30"
    _0145 = "01:45"
    _0200 = "02:00"
    _0215 = "02:15"
    _0230 = "02:30"
    _0245 = "02:45"
    _0300 = "03:00"
    _0315 = "03:15"
    _0330 = "03:30"
    _0345 = "03:45"
    _0400 = "04:00"
    _0415 = "04:15"
    _0430 = "04:30"
    _0445 = "04:45"
    _0500 = "05:00"
    _0515 = "05:15"
    _0530 = "05:30"
    _0545 = "05:45"
    _0600 = "06:00"
    _0615 = "06:15"
    _0630 = "06:30"
    _0645 = "06:45"
    _0700 = "07:00"
    _0715 = "07:15"
    _0730 = "07:30"
    _0745 = "07:45"
    _0800 = "08:00"
    _0815 = "08:15"
    _0830 = "08:30"
    _0845 = "08:45"
    _0900 = "09:00"
    _0915 = "09:15"
    _0930 = "09:30"
    _0945 = "09:45"
    _1000 = "10:00"
    _1015 = "10:15"
    _1030 = "10:30"
    _1045 = "10:45"
    _1100 = "11:00"
    _1115 = "11:15"
    _1130 = "11:30"
    _1145 = "11:45"
    _1200 = "12:00"
    _1215 = "12:15"
    _1230 = "12:30"
    _1245 = "12:45"
    _1300 = "13:00"
    _1315 = "13:15"
    _1330 = "13:30"
    _1345 = "13:45"
    _1400 = "14:00"
    _1415 = "14:15"
    _1430 = "14:30"
    _1445 = "14:45"
    _1500 = "15:00"
    _1515 = "15:15"
    _1530 = "15:30"
    _1545 = "15:45"
    _1600 = "16:00"
    _1615 = "16:15"
    _1630 = "16:30"
    _1645 = "16:45"
    _1700 = "17:00"
    _1715 = "17:15"
    _1730 = "17:30"
    _1745 = "17:45"
    _1800 = "18:00"
    _1815 = "18:15"
    _1830 = "18:30"
    _1845 = "18:45"
    _1900 = "19:00"
    _1915 = "19:15"
    _1930 = "19:30"
    _1945 = "19:45"
    _2000 = "20:00"
    _2015 = "20:15"
    _2030 = "20:30"
    _2045 = "20:45"
    _2100 = "21:00"
    _2115 = "21:15"
    _2130 = "21:30"
    _2145 = "21:45"
    _2200 = "22:00"
    _2215 = "22:15"
    _2230 = "22:30"
    _2245 = "22:45"
    _2300 = "23:00"
    _2315 = "23:15"
    _2330 = "23:30"
    _2345 = "23:45"

    def to_time(self) -> time:
        return datetime.strptime(self.value, "%H:%M").time()


class EventVisibility(Enum):
    DEFAULT = "default"
    PUBLIC = "public"
    PRIVATE = "private"
    CONFIDENTIAL = "confidential"


class EventType(Enum):
    BIRTHDAY = "birthday"  # Special all-day events with an annual recurrence.
    DEFAULT = "default"  # Regular events
    FOCUS_TIME = "focusTime"  # Focus time events
    FROM_GMAIL = "fromGmail"  # Events from Gmail
    OUT_OF_OFFICE = "outOfOffice"  # Out of office events
    WORKING_LOCATION = "workingLocation"  # Working location events


class SendUpdatesOptions(Enum):
    NONE = "none"  # No notifications are sent
    ALL = "all"  # Notifications are sent to all guests
    EXTERNAL_ONLY = "externalOnly"  # Notifications are sent to non-Google Calendar guests only.


# ---------------------------------------------------------------------------- #
# Google Drive Models and Enums
# ---------------------------------------------------------------------------- #
class Corpora(str, Enum):
    """
    Bodies of items (files/documents) to which the query applies.
    Prefer 'user' or 'drive' to 'allDrives' for efficiency.
    By default, corpora is set to 'user'.
    """

    USER = "user"
    DOMAIN = "domain"
    DRIVE = "drive"
    ALL_DRIVES = "allDrives"


class OrderBy(str, Enum):
    """
    Sort keys for ordering files in Google Drive.
    Each key has both ascending and descending options.
    """

    CREATED_TIME = (
        # When the file was created (ascending)
        "createdTime"
    )
    CREATED_TIME_DESC = (
        # When the file was created (descending)
        "createdTime desc"
    )
    FOLDER = (
        # The folder ID, sorted using alphabetical ordering (ascending)
        "folder"
    )
    FOLDER_DESC = (
        # The folder ID, sorted using alphabetical ordering (descending)
        "folder desc"
    )
    MODIFIED_BY_ME_TIME = (
        # The last time the file was modified by the user (ascending)
        "modifiedByMeTime"
    )
    MODIFIED_BY_ME_TIME_DESC = (
        # The last time the file was modified by the user (descending)
        "modifiedByMeTime desc"
    )
    MODIFIED_TIME = (
        # The last time the file was modified by anyone (ascending)
        "modifiedTime"
    )
    MODIFIED_TIME_DESC = (
        # The last time the file was modified by anyone (descending)
        "modifiedTime desc"
    )
    NAME = (
        # The name of the file, sorted using alphabetical ordering (e.g., 1, 12, 2, 22) (ascending)
        "name"
    )
    NAME_DESC = (
        # The name of the file, sorted using alphabetical ordering (e.g., 1, 12, 2, 22) (descending)
        "name desc"
    )
    NAME_NATURAL = (
        # The name of the file, sorted using natural sort ordering (e.g., 1, 2, 12, 22) (ascending)
        "name_natural"
    )
    NAME_NATURAL_DESC = (
        # The name of the file, sorted using natural sort ordering (e.g., 1, 2, 12, 22) (descending)
        "name_natural desc"
    )
    QUOTA_BYTES_USED = (
        # The number of storage quota bytes used by the file (ascending)
        "quotaBytesUsed"
    )
    QUOTA_BYTES_USED_DESC = (
        # The number of storage quota bytes used by the file (descending)
        "quotaBytesUsed desc"
    )
    RECENCY = (
        # The most recent timestamp from the file's date-time fields (ascending)
        "recency"
    )
    RECENCY_DESC = (
        # The most recent timestamp from the file's date-time fields (descending)
        "recency desc"
    )
    SHARED_WITH_ME_TIME = (
        # When the file was shared with the user, if applicable (ascending)
        "sharedWithMeTime"
    )
    SHARED_WITH_ME_TIME_DESC = (
        # When the file was shared with the user, if applicable (descending)
        "sharedWithMeTime desc"
    )
    STARRED = (
        # Whether the user has starred the file (ascending)
        "starred"
    )
    STARRED_DESC = (
        # Whether the user has starred the file (descending)
        "starred desc"
    )
    VIEWED_BY_ME_TIME = (
        # The last time the file was viewed by the user (ascending)
        "viewedByMeTime"
    )
    VIEWED_BY_ME_TIME_DESC = (
        # The last time the file was viewed by the user (descending)
        "viewedByMeTime desc"
    )


class DocumentFormat(str, Enum):
    MARKDOWN = "markdown"
    HTML = "html"
    GOOGLE_API_JSON = "google_api_json"


# ---------------------------------------------------------------------------- #
# Google Gmail Models and Enums
# ---------------------------------------------------------------------------- #
class GmailReplyToWhom(str, Enum):
    EVERY_RECIPIENT = "every_recipient"
    ONLY_THE_SENDER = "only_the_sender"


class GmailAction(str, Enum):
    SEND = "send"
    DRAFT = "draft"


# ---------------------------------------------------------------------------- #
# Google Sheets Models and Enums
# ---------------------------------------------------------------------------- #
class CellErrorType(str, Enum):
    """The type of error in a cell

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ErrorType
    """

    ERROR_TYPE_UNSPECIFIED = "ERROR_TYPE_UNSPECIFIED"  # The default error type, do not use this.
    ERROR = "ERROR"  # Corresponds to the #ERROR! error.
    NULL_VALUE = "NULL_VALUE"  # Corresponds to the #NULL! error.
    DIVIDE_BY_ZERO = "DIVIDE_BY_ZERO"  # Corresponds to the #DIV/0 error.
    VALUE = "VALUE"  # Corresponds to the #VALUE! error.
    REF = "REF"  # Corresponds to the #REF! error.
    NAME = "NAME"  # Corresponds to the #NAME? error.
    NUM = "NUM"  # Corresponds to the #NUM! error.
    N_A = "N_A"  # Corresponds to the #N/A error.
    LOADING = "LOADING"  # Corresponds to the Loading... state.


class CellErrorValue(BaseModel):
    """An error in a cell

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ErrorValue
    """

    type: CellErrorType
    message: str


class CellExtendedValue(BaseModel):
    """The kinds of value that a cell in a spreadsheet can have

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ExtendedValue
    """

    numberValue: float | None = None
    stringValue: str | None = None
    boolValue: bool | None = None
    formulaValue: str | None = None
    errorValue: Optional["CellErrorValue"] = None

    @model_validator(mode="after")
    def check_exactly_one_value(cls, instance):  # type: ignore[no-untyped-def]
        provided = [v for v in instance.__dict__.values() if v is not None]
        if len(provided) != 1:
            raise ValueError(
                "Exactly one of numberValue, stringValue, boolValue, "
                "formulaValue, or errorValue must be set."
            )
        return instance


class NumberFormatType(str, Enum):
    NUMBER = "NUMBER"
    PERCENT = "PERCENT"
    CURRENCY = "CURRENCY"


class NumberFormat(BaseModel):
    """The format of a number

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#NumberFormat
    """

    pattern: str
    type: NumberFormatType


class CellFormat(BaseModel):
    """The format of a cell

    Partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellFormat
    """

    numberFormat: NumberFormat


class CellData(BaseModel):
    """Data about a specific cell

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData
    """

    userEnteredValue: CellExtendedValue
    userEnteredFormat: CellFormat | None = None


class RowData(BaseModel):
    """Data about each cellin a row

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#RowData
    """

    values: list[CellData]


class GridData(BaseModel):
    """Data in the grid

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#GridData
    """

    startRow: int
    startColumn: int
    rowData: list[RowData]


class GridProperties(BaseModel):
    """Properties of a grid

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#GridProperties
    """

    rowCount: int
    columnCount: int


class SheetProperties(BaseModel):
    """Properties of a Sheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#SheetProperties
    """

    sheetId: int
    title: str
    gridProperties: GridProperties | None = None


class Sheet(BaseModel):
    """A Sheet in a spreadsheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#Sheet
    """

    properties: SheetProperties
    data: list[GridData] | None = None


class SpreadsheetProperties(BaseModel):
    """Properties of a spreadsheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#SpreadsheetProperties
    """

    title: str


class Spreadsheet(BaseModel):
    """A spreadsheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets
    """

    properties: SpreadsheetProperties
    sheets: list[Sheet]


CellValue = int | float | str | bool


class SheetDataInput(BaseModel):
    """
    SheetDataInput models the cell data of a spreadsheet in a custom format.

    It is a dictionary mapping row numbers (as ints) to dictionaries that map
    column letters (as uppercase strings) to cell values (int, float, str, or bool).

    This model enforces that:
      - The outer keys are convertible to int.
      - The inner keys are alphabetic strings (normalized to uppercase).
      - All cell values are only of type int, float, str, or bool.

    The model automatically serializes (via `json_data()`)
    and validates the inner types.
    """

    data: dict[int, dict[str, CellValue]]

    @classmethod
    def _parse_json_if_string(cls, value):  # type: ignore[no-untyped-def]
        """Parses the value if it is a JSON string, otherwise returns it.

        Helper method for when validating the `data` field.
        """
        if isinstance(value, str):
            try:
                return json.loads(value)
            except json.JSONDecodeError as e:
                raise TypeError(f"Invalid JSON: {e}")
        return value

    @classmethod
    def _validate_row_key(cls, row_key) -> int:  # type: ignore[no-untyped-def]
        """Converts the row key to an integer, raising an error if conversion fails.

        Helper method for when validating the `data` field.
        """
        try:
            return int(row_key)
        except (ValueError, TypeError):
            raise TypeError(f"Row key '{row_key}' is not convertible to int.")

    @classmethod
    def _validate_inner_cells(cls, cells, row_int: int) -> dict:  # type: ignore[no-untyped-def]
        """Validates that 'cells' is a dict mapping column letters to valid cell values
        and normalizes the keys.

        Helper method for when validating the `data` field.
        """
        if not isinstance(cells, dict):
            raise TypeError(
                f"Value for row '{row_int}' must be a dict mapping column letters to cell values."
            )
        new_inner = {}
        for col_key, cell_value in cells.items():
            if not isinstance(col_key, str):
                raise TypeError(f"Column key '{col_key}' must be a string.")
            col_string = col_key.upper()
            if not col_string.isalpha():
                raise TypeError(f"Column key '{col_key}' is invalid. Must be alphabetic.")
            if not isinstance(cell_value, int | float | str | bool):
                raise TypeError(
                    f"Cell value for {col_string}{row_int} must be an int, float, str, or bool."
                )
            new_inner[col_string] = cell_value
        return new_inner

    @field_validator("data", mode="before")
    @classmethod
    def validate_and_convert_keys(cls, value):  # type: ignore[no-untyped-def]
        """
        Validates data when SheetDataInput is instantiated and converts it to the correct format.
        Uses private helper methods to parse JSON, validate row keys, and validate inner cell data.
        """
        if value is None:
            return {}

        value = cls._parse_json_if_string(value)
        if isinstance(value, dict):
            new_value = {}
            for row_key, cells in value.items():
                row_int = cls._validate_row_key(row_key)
                inner_cells = cls._validate_inner_cells(cells, row_int)
                new_value[row_int] = inner_cells
            return new_value

        raise TypeError("data must be a dict or a valid JSON string representing a dict")

    def json_data(self) -> str:
        """
        Serialize the sheet data to a JSON string.
        """
        return json.dumps(self.data)

    @classmethod
    def from_json(cls, json_str: str) -> "SheetDataInput":
        """
        Create a SheetData instance from a JSON string.
        """
        return cls.model_validate_json(json_str)



import httpx
from arcade_tdk import ToolContext

from arcade_spotify.tools.constants import ENDPOINTS, SPOTIFY_BASE_URL
from arcade_spotify.tools.models import PlaybackState


async def send_spotify_request(
    context: ToolContext,
    method: str,
    url: str,
    params: dict | None = None,
    json_data: dict | None = None,
) -> httpx.Response:
    """
    Send an asynchronous request to the Spotify API.

    Args:
        context: The tool context containing the authorization token.
        method: The HTTP method (GET, POST, PUT, DELETE, etc.).
        url: The full URL for the API endpoint.
        params: Query parameters to include in the request.
        json_data: JSON data to include in the request body.

    Returns:
        The response object from the API request.

    Raises:
        ToolExecutionError: If the request fails for any reason.
    """
    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    headers = {"Authorization": f"Bearer {token}"}

    async with httpx.AsyncClient() as client:
        response = await client.request(method, url, headers=headers, params=params, json=json_data)

    return response


def get_url(endpoint: str, **kwargs: object) -> str:
    """
    Get the full Spotify URL for a given endpoint.

    :param endpoint: The endpoint key from ENDPOINTS
    :param kwargs: The parameters to format the URL with
    :return: The full URL
    """
    return f"{SPOTIFY_BASE_URL}{ENDPOINTS[endpoint].format(**kwargs)}"


def convert_to_playback_state(data: dict) -> PlaybackState:
    """
    Convert the Spotify API endpoint "/me/player" response data to a PlaybackState object.

    Args:
        data: The response data from the Spotify API endpoint "/me/player".

    Returns:
        An instance of PlaybackState populated with the data.
    """
    playback_state = PlaybackState(
        device_name=data.get("device", {}).get("name"),
        device_id=data.get("device", {}).get("id"),
        currently_playing_type=data.get("currently_playing_type"),
        is_playing=data.get("is_playing"),
        progress_ms=data.get("progress_ms"),
        message=data.get("message"),
    )

    if data.get("currently_playing_type") == "track":
        item = data.get("item") or {}
        album = item.get("album", {})
        playback_state.album_name = album.get("name")
        playback_state.album_id = album.get("id")
        playback_state.album_artists = [artist.get("name") for artist in album.get("artists", [])]
        playback_state.album_spotify_url = album.get("external_urls", {}).get("spotify")
        playback_state.track_name = item.get("name")
        playback_state.track_id = item.get("id")
        playback_state.track_spotify_url = item.get("external_urls", {}).get("spotify")
        playback_state.track_artists = [artist.get("name") for artist in item.get("artists", [])]
        playback_state.track_artists_ids = [artist.get("id") for artist in item.get("artists", [])]
    elif data.get("currently_playing_type") == "episode":
        item = data.get("item") or {}
        show = item.get("show", {})
        playback_state.show_name = show.get("name")
        playback_state.show_id = show.get("id")
        playback_state.show_spotify_url = show.get("external_urls", {}).get("spotify")
        playback_state.episode_name = item.get("name")
        playback_state.episode_id = item.get("id")
        playback_state.episode_spotify_url = item.get("external_urls", {}).get("spotify")

    return playback_state



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_outlook_mail import create_draft_email, update_draft_email
from evals.additional_messages import (
    update_draft_email_additional_messages,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_tool(create_draft_email, "OutlookMail")
catalog.add_tool(update_draft_email, "OutlookMail")


@tool_eval()
def outlook_mail_write_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Outlook Mail tools."""
    suite = EvalSuite(
        name="Outlook Mail Write Evaluation",
        system_message=("You are an AI that has access to tools to send, read, and write emails."),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Create draft email",
        user_message=(
            "create a new draft email with subject 'Hello friends' and body "
            "'I've gathered you all here to celebrate the launch of the new Arcade platform."
            "address it to e@arcade.dev and z@arcade.dev. also carbon copy to j@arcade.dev, "
            "f@arcade.dev, k@arcade.dev and finally to m@arcade.dev. also bcc to r@arcade.dev"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_draft_email,
                args={
                    "subject": "Hello friends",
                    "body": "I've gathered you all here to celebrate the launch of the new Arcade platform.",
                    "to_recipients": ["e@arcade.dev", "z@arcade.dev"],
                    "cc_recipients": [
                        "j@arcade.dev",
                        "f@arcade.dev",
                        "k@arcade.dev",
                        "m@arcade.dev",
                    ],
                    "bcc_recipients": ["r@arcade.dev"],
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=0.2),
            SimilarityCritic(critic_field="body", weight=0.2),
            BinaryCritic(critic_field="to_recipients", weight=0.2),
            BinaryCritic(critic_field="cc_recipients", weight=0.2),
            BinaryCritic(critic_field="bcc_recipients", weight=0.2),
        ],
    )

    suite.add_case(
        name="Update draft email",
        user_message=(
            "oh wait i think i messed up on some emails. I meant 'z', not 'e'. "
            "Also, I forgot to bcc y@arcade.dev. Also, replace the period with an "
            "exclamation point since I want to convey excitement. Oh I almost forgot, "
            "Don't cc anyone."
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=update_draft_email,
                args={
                    "message_id": "AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDwAAAFuxokOLZRtDncM4_x_WeUwAAAAC-dpvAAAA",
                    "body": "I've gathered you all here to celebrate the launch of the new Arcade platform!",
                    "to_add": ["z@arcade.dev"],
                    "to_remove": ["e@arcade.dev"],
                    "cc_remove": ["j@arcade.dev", "f@arcade.dev", "k@arcade.dev", "m@arcade.dev"],
                    "bcc_add": ["y@arcade.dev"],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="message_id", weight=1 / 6),
            BinaryCritic(critic_field="body", weight=1 / 6),
            BinaryCritic(critic_field="to_add", weight=1 / 6),
            BinaryCritic(critic_field="to_remove", weight=1 / 6),
            BinaryCritic(critic_field="cc_remove", weight=1 / 6),
            BinaryCritic(critic_field="bcc_add", weight=1 / 6),
        ],
        additional_messages=update_draft_email_additional_messages,
    )

    return suite



import asyncio
import json
import json.decoder
from dataclasses import dataclass
from typing import Any, cast

import httpx

import arcade_jira.cache as cache
from arcade_jira.constants import JIRA_API_VERSION, JIRA_BASE_URL, JIRA_MAX_CONCURRENT_REQUESTS
from arcade_jira.exceptions import JiraToolExecutionError, NotFoundError


@dataclass
class JiraClient:
    auth_token: str
    base_url: str = JIRA_BASE_URL
    api_version: str = JIRA_API_VERSION
    max_concurrent_requests: int = JIRA_MAX_CONCURRENT_REQUESTS
    _semaphore: asyncio.Semaphore | None = None
    _cloud_id: str | None = None

    def __post_init__(self) -> None:
        if not self._semaphore:
            cached_semaphore = cache.get_jira_client_semaphore(self.auth_token)

            if cached_semaphore:
                self._semaphore = cached_semaphore
            else:
                self._semaphore = asyncio.Semaphore(self.max_concurrent_requests)
                cache.set_jira_client_semaphore(self.auth_token, self._semaphore)

        self.base_url = self.base_url.rstrip("/")
        self.api_version = self.api_version.strip("/")

    async def get_cloud_id(self) -> str:
        if self._cloud_id is None:
            if (cloud_id := await cache.async_get_cloud_id(self.auth_token)) is not None:
                self._cloud_id = cloud_id
            else:
                cloud = await self._get_cloud_data_from_available_resources()
                self._cloud_id = cloud["id"]
                await cache.async_set_cloud_id(self.auth_token, cloud["id"])
                await cache.async_set_cloud_name(self.auth_token, cloud["name"])

        return self._cloud_id

    async def _build_url(self, endpoint: str) -> str:
        cloud_id = await self.get_cloud_id()
        return f"{self.base_url}/{cloud_id}/rest/api/{self.api_version}/{endpoint.lstrip('/')}"

    async def _get_cloud_data_from_available_resources(self) -> dict[str, Any]:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://api.atlassian.com/oauth/token/accessible-resources",
                headers={"Authorization": f"Bearer {self.auth_token}"},
            )

            available_resources = deduplicate_available_resources(response.json())

            if len(available_resources) == 0:
                raise JiraToolExecutionError(
                    message="No cloud ID returned by Atlassian, cannot make API calls"
                )
            if len(available_resources) > 1:
                cloud_ids_found = json.dumps([
                    {
                        "id": resource["id"],
                        "name": resource["name"],
                        "url": resource["url"],
                    }
                    for resource in available_resources
                ])
                raise JiraToolExecutionError(
                    message=(
                        "Multiple cloud IDs returned by Atlassian, cannot resolve which one "
                        "to use. Please revoke your authorization access and authorize a single "
                        f"Atlassian Cloud. Available cloud IDs: {cloud_ids_found}. "
                    )
                )
            return cast(dict[str, Any], available_resources[0])

    def _build_error_messages(self, response: httpx.Response) -> tuple[str, str | None]:
        try:
            data = response.json()
            developer_message = None

            if "errorMessages" in data:
                if len(data["errorMessages"]) == 1:
                    error_message = cast(str, data["errorMessages"][0])
                elif "errors" in data:
                    error_message = json.dumps(data["errors"])
                else:
                    error_message = "Unknown error"

            elif "message" in data:
                error_message = cast(str, data["message"])

            else:
                error_message = json.dumps(data)

        except Exception as e:
            error_message = "Failed to parse Jira error response"
            developer_message = (
                f"Failed to parse Jira error response: {type(e).__name__}: {e!s}. "
                f"API Response: {response.text}"
            )

        return error_message, developer_message

    def _raise_for_status(self, response: httpx.Response) -> None:
        if response.status_code < 300:
            return

        error_message, developer_message = self._build_error_messages(response)

        if response.status_code == 404:
            raise NotFoundError(error_message, developer_message)

        raise JiraToolExecutionError(error_message, developer_message)

    def _set_request_body(self, kwargs: dict, data: dict | None, json_data: dict | None) -> dict:
        if data and json_data:
            raise ValueError("Cannot provide both data and json_data")  # noqa: TRY003

        if data:
            kwargs["data"] = data

        elif json_data:
            kwargs["json"] = json_data

        return kwargs

    def _format_response_dict(self, response: httpx.Response) -> dict:
        try:
            return cast(dict, response.json())
        except (UnicodeDecodeError, json.decoder.JSONDecodeError):
            return {"text": response.text}

    async def get(
        self,
        endpoint: str,
        params: dict | None = None,
        headers: dict | None = None,
    ) -> dict:
        default_headers = {
            "Authorization": f"Bearer {self.auth_token}",
            "Accept": "application/json",
        }
        headers = {**default_headers, **(headers or {})}

        kwargs = {
            "url": await self._build_url(endpoint),
            "headers": headers,
        }

        if params:
            kwargs["params"] = params

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.get(**kwargs)  # type: ignore[arg-type]
            self._raise_for_status(response)

        return self._format_response_dict(response)

    async def post(
        self,
        endpoint: str,
        data: dict | None = None,
        json_data: dict | None = None,
        files: dict | None = None,
        headers: dict | None = None,
    ) -> dict:
        default_headers = {
            "Authorization": f"Bearer {self.auth_token}",
            "Accept": "application/json",
        }

        if files is None and json_data is not None:
            default_headers["Content-Type"] = "application/json"

        headers = {**default_headers, **(headers or {})}

        kwargs = {
            "url": await self._build_url(endpoint),
            "headers": headers,
        }

        if files is not None:
            kwargs["files"] = files
            if data is not None:
                kwargs["data"] = data
        else:
            kwargs = self._set_request_body(kwargs, data, json_data)

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.post(**kwargs)  # type: ignore[arg-type]
            self._raise_for_status(response)

        return self._format_response_dict(response)

    async def put(
        self,
        endpoint: str,
        data: dict | None = None,
        json_data: dict | None = None,
        params: dict | None = None,
        headers: dict | None = None,
    ) -> dict:
        headers = headers or {}
        headers["Authorization"] = f"Bearer {self.auth_token}"
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"

        kwargs = {
            "url": await self._build_url(endpoint),
            "headers": headers,
        }

        kwargs = self._set_request_body(kwargs, data, json_data)

        if params:
            kwargs["params"] = params

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.put(**kwargs)  # type: ignore[arg-type]
            self._raise_for_status(response)

        return self._format_response_dict(response)


def deduplicate_available_resources(available_resources: list[dict]) -> list[dict]:
    account_ids_seen = set()
    deduplicated = []

    for item in available_resources:
        if item["id"] not in account_ids_seen:
            deduplicated.append(item)
            account_ids_seen.add(item["id"])

    return deduplicated



from typing import Annotated, Optional

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Dropbox
from arcade_tdk.errors import ToolExecutionError

from arcade_dropbox.constants import Endpoint, ItemCategory
from arcade_dropbox.exceptions import DropboxApiError
from arcade_dropbox.utils import (
    build_dropbox_json,
    clean_dropbox_entries,
    parse_dropbox_path,
    send_dropbox_request,
)


@tool(
    requires_auth=Dropbox(
        scopes=["files.metadata.read"],
    )
)
async def list_items_in_folder(
    context: ToolContext,
    folder_path: Annotated[
        str,
        "The path to the folder to list the contents of. E.g. '/AcmeInc/Reports'. "
        "Defaults to an empty string (list items in the Dropbox root folder).",
    ] = "",
    limit: Annotated[
        int,
        "The maximum number of items to return. Defaults to 100. Maximum allowed is 2000.",
    ] = 100,
    cursor: Annotated[
        Optional[str],
        "The cursor token for the next page of results. "
        "Defaults to None (returns the first page of results).",
    ] = None,
) -> Annotated[
    dict, "Dictionary containing the list of files and folders in the specified folder path"
]:
    """Provides a dictionary containing the list of items in the specified folder path.

    Note 1: when paginating, it is not necessary to provide any other argument besides the cursor.
    Note 2: when paginating, any given item (file or folder) may be returned in multiple pages.
    """
    limit = min(limit, 2000)

    try:
        result = await send_dropbox_request(
            context.get_auth_token_or_empty(),
            endpoint=Endpoint.LIST_FOLDER,
            path=parse_dropbox_path(folder_path),
            limit=limit,
            cursor=cursor,
        )
    except DropboxApiError as api_error:
        return {"error": api_error.message}

    return {
        "items": clean_dropbox_entries(result["entries"]),
        "cursor": result.get("cursor"),
        "has_more": result.get("has_more", False),
    }


@tool(
    requires_auth=Dropbox(
        scopes=["files.metadata.read"],
    )
)
async def search_files_and_folders(
    context: ToolContext,
    keywords: Annotated[
        str,
        "The keywords to search for. E.g. 'quarterly report'. "
        "Maximum length allowed by the Dropbox API is 1000 characters. ",
    ],
    search_in_folder_path: Annotated[
        Optional[str],
        "Restricts the search to the specified folder path. E.g. '/AcmeInc/Reports'. "
        "Defaults to None (search in the entire Dropbox).",
    ] = None,
    filter_by_category: Annotated[
        Optional[list[ItemCategory]],
        "Restricts the search to the specified category(ies) of items. "
        "Provide None, one or multiple, if needed. Defaults to None (returns all categories).",
    ] = None,
    limit: Annotated[
        int,
        "The maximum number of items to return. Defaults to 100. Maximum allowed is 1000.",
    ] = 100,
    cursor: Annotated[
        Optional[str],
        "The cursor token for the next page of results. Defaults to None (first page of results).",
    ] = None,
) -> Annotated[dict, "List of items in the specified folder path matching the search criteria"]:
    """Returns a list of items in the specified folder path matching the search criteria.

    Note 1: the Dropbox API will return up to 10,000 (ten thousand) items cumulatively across
    multiple pagination requests using the cursor token.
    Note 2: when paginating, it is not necessary to provide any other argument besides the cursor.
    Note 3: when paginating, any given item (file or folder) may be returned in multiple pages.
    """
    if len(keywords) > 1000:
        raise ToolExecutionError(
            "The keywords argument must be a string with up to 1000 characters."
        )

    limit = min(limit, 1000)

    filter_by_category = filter_by_category or []

    options = build_dropbox_json(
        file_status="active",
        filename_only=False,
        path=parse_dropbox_path(search_in_folder_path),
        max_results=limit,
        file_categories=[category.value for category in filter_by_category],
    )

    try:
        result = await send_dropbox_request(
            context.get_auth_token_or_empty(),
            endpoint=Endpoint.SEARCH_FILES,
            query=keywords,
            options=options,
            cursor=cursor,
        )
    except DropboxApiError as api_error:
        return {"error": api_error.message}

    return {
        "items": clean_dropbox_entries([
            match["metadata"]["metadata"] for match in result["matches"]
        ]),
        "cursor": result.get("cursor"),
        "has_more": result.get("has_more", False),
    }



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from arcade_tdk.errors import RetryableToolError
from sqlalchemy import inspect, text
from sqlalchemy.ext.asyncio import AsyncEngine

from ..database_engine import MAX_ROWS_RETURNED, DatabaseEngine


@tool(requires_secrets=["DATABASE_CONNECTION_STRING"])
async def discover_schemas(
    context: ToolContext,
) -> list[str]:
    """Discover all the schemas in the postgres database."""
    async with await DatabaseEngine.get_engine(
        context.get_secret("DATABASE_CONNECTION_STRING")
    ) as engine:
        schemas = await _get_schemas(engine)
        return schemas


@tool(requires_secrets=["DATABASE_CONNECTION_STRING"])
async def discover_tables(
    context: ToolContext,
    schema_name: Annotated[
        str, "The database schema to discover tables in (default value: 'public')"
    ] = "public",
) -> list[str]:
    """Discover all the tables in the postgres database when the list of tables is not known.

    THIS TOOL SHOULD ALWAYS BE USED BEFORE ANY OTHER TOOL THAT REQUIRES A TABLE NAME.
    """
    async with await DatabaseEngine.get_engine(
        context.get_secret("DATABASE_CONNECTION_STRING")
    ) as engine:
        tables = await _get_tables(engine, schema_name)
        return tables


@tool(requires_secrets=["DATABASE_CONNECTION_STRING"])
async def get_table_schema(
    context: ToolContext,
    schema_name: Annotated[str, "The database schema to get the table schema of"],
    table_name: Annotated[str, "The table to get the schema of"],
) -> list[str]:
    """
    Get the schema/structure of a postgres table in the postgres database when the schema is not known, and the name of the table is provided.

    THIS TOOL SHOULD ALWAYS BE USED BEFORE EXECUTING ANY QUERY.  ALL TABLES IN THE QUERY MUST BE DISCOVERED FIRST USING THE <DiscoverTables> TOOL.
    """
    async with await DatabaseEngine.get_engine(
        context.get_secret("DATABASE_CONNECTION_STRING")
    ) as engine:
        return await _get_table_schema(engine, schema_name, table_name)


@tool(requires_secrets=["DATABASE_CONNECTION_STRING"])
async def execute_query(
    context: ToolContext,
    query: Annotated[str, "The postgres SQL query to execute.  Only SELECT queries are allowed."],
) -> list[str]:
    """
    You have a connection to a postgres database.
    Execute a query and return the results against the postgres database.

    ONLY USE THIS TOOL IF YOU HAVE ALREADY LOADED THE SCHEMA OF THE TABLES YOU NEED TO QUERY.  USE THE <GetTableSchema> TOOL TO LOAD THE SCHEMA IF NOT ALREADY KNOWN.

    When running queries, follow these rules which will help avoid errors:
    * Always use case-insensitive queries to match strings in the query.
    * Always trim strings in the query.
    * Prefer LIKE queries over direct string matches or regex queries.
    * Only join on columns that are indexed or the primary key.  Do not join on arbitrary columns.

    Only SELECT queries are allowed.  Do not use INSERT, UPDATE, DELETE, or other DML statements.  This tool will reject them.

    Unless otherwise specified, ensure that query has a LIMIT of 100 for all results.  This tool will enforce that no more than 1000 rows are returned at maximum.
    """
    async with await DatabaseEngine.get_engine(
        context.get_secret("DATABASE_CONNECTION_STRING")
    ) as engine:
        try:
            return await _execute_query(engine, query)
        except Exception as e:
            raise RetryableToolError(
                f"Query failed: {e}",
                developer_message=f"Query '{query}' failed.",
                additional_prompt_content="Load the database schema <GetTableSchema> or use the <DiscoverTables> tool to discover the tables and try again.",
                retry_after_ms=10,
            ) from e


async def _get_schemas(engine: AsyncEngine) -> list[str]:
    """Get all the schemas in the database"""
    async with engine.connect() as conn:

        def get_schema_names(sync_conn: Any) -> list[str]:
            return list(inspect(sync_conn).get_schema_names())

        schemas: list[str] = await conn.run_sync(get_schema_names)
        schemas = [schema for schema in schemas if schema != "information_schema"]

        return schemas


async def _get_tables(engine: AsyncEngine, schema_name: str) -> list[str]:
    """Get all the tables in the database"""
    async with engine.connect() as conn:

        def get_schema_names(sync_conn: Any) -> list[str]:
            return list(inspect(sync_conn).get_schema_names())

        schemas: list[str] = await conn.run_sync(get_schema_names)
        tables = []
        for schema in schemas:
            if schema == schema_name:

                def get_table_names(sync_conn: Any, s: str = schema) -> list[str]:
                    return list(inspect(sync_conn).get_table_names(schema=s))

                these_tables = await conn.run_sync(get_table_names)
                tables.extend(these_tables)
        return tables


async def _get_table_schema(engine: AsyncEngine, schema_name: str, table_name: str) -> list[str]:
    """Get the schema of a table"""
    async with engine.connect() as connection:

        def get_columns(sync_conn: Any, t: str = table_name, s: str = schema_name) -> list[Any]:
            return list(inspect(sync_conn).get_columns(t, s))

        columns_table = await connection.run_sync(get_columns)

        # Get primary key information
        pk_constraint = await connection.run_sync(
            lambda sync_conn: inspect(sync_conn).get_pk_constraint(table_name, schema_name)
        )
        primary_keys = set(pk_constraint.get("constrained_columns", []))

        # Get index information
        indexes = await connection.run_sync(
            lambda sync_conn: inspect(sync_conn).get_indexes(table_name, schema_name)
        )
        indexed_columns = set()
        for index in indexes:
            indexed_columns.update(index.get("column_names", []))

        results = []
        for column in columns_table:
            column_name = column["name"]
            column_type = column["type"].python_type.__name__

            # Build column description
            description = f"{column_name}: {column_type}"

            # Add primary key indicator
            if column_name in primary_keys:
                description += " (PRIMARY KEY)"

            # Add index indicator
            if column_name in indexed_columns:
                description += " (INDEXED)"

            results.append(description)

        return results[:MAX_ROWS_RETURNED]


async def _execute_query(
    engine: AsyncEngine, query: str, params: dict[str, Any] | None = None
) -> list[str]:
    """Execute a query and return the results."""
    async with engine.connect() as connection:
        result = await connection.execute(text(DatabaseEngine.sanitize_query(query)), params)
        rows = result.fetchall()
        results = [str(row) for row in rows]
        return results[:MAX_ROWS_RETURNED]



# Maps "Windows timezone format" to "IANA timezone format"
# Does not include all Windows timezones.
WINDOWS_TO_IANA = {
    "Dateline Standard Time": "Etc/GMT+12",
    "UTC-11": "Etc/GMT+11",
    "Aleutian Standard Time": "America/Adak",
    "Hawaiian Standard Time": "Pacific/Honolulu",
    "Marquesas Standard Time": "Pacific/Marquesas",
    "Alaskan Standard Time": "America/Anchorage",
    "UTC-09": "Etc/GMT+9",
    "Pacific Standard Time (Mexico)": "America/Tijuana",
    "UTC-08": "Etc/GMT+8",
    "Pacific Standard Time": "America/Los_Angeles",
    "US Mountain Standard Time": "America/Phoenix",
    "Mountain Standard Time (Mexico)": "America/Chihuahua",
    "Mountain Standard Time": "America/Denver",
    "Central America Standard Time": "America/Guatemala",
    "Central Standard Time": "America/Chicago",
    "Easter Island Standard Time": "Pacific/Easter",
    "Central Standard Time (Mexico)": "America/Mexico_City",
    "Canada Central Standard Time": "America/Regina",
    "SA Pacific Standard Time": "America/Bogota",
    "Eastern Standard Time (Mexico)": "America/Cancun",
    "Eastern Standard Time": "America/New_York",
    "Haiti Standard Time": "America/Port-au-Prince",
    "Cuba Standard Time": "America/Havana",
    "US Eastern Standard Time": "America/Indianapolis",
    "Turks And Caicos Standard Time": "America/Grand_Turk",
    "Paraguay Standard Time": "America/Asuncion",
    "Atlantic Standard Time": "America/Halifax",
    "Venezuela Standard Time": "America/Caracas",
    "Central Brazilian Standard Time": "America/Cuiaba",
    "SA Western Standard Time": "America/La_Paz",
    "Pacific SA Standard Time": "America/Santiago",
    "Newfoundland Standard Time": "America/St_Johns",
    "Tocantins Standard Time": "America/Araguaina",
    "E. South America Standard Time": "America/Sao_Paulo",
    "SA Eastern Standard Time": "America/Cayenne",
    "Argentina Standard Time": "America/Buenos_Aires",
    "Greenland Standard Time": "America/Godthab",
    "Montevideo Standard Time": "America/Montevideo",
    "Magallanes Standard Time": "America/Punta_Arenas",
    "Saint Pierre Standard Time": "America/Miquelon",
    "Bahia Standard Time": "America/Bahia",
    "UTC-02": "Etc/GMT+2",
    "Azores Standard Time": "Atlantic/Azores",
    "Cape Verde Standard Time": "Atlantic/Cape_Verde",
    "UTC": "Etc/UTC",
    "GMT Standard Time": "Europe/London",
    "Greenwich Standard Time": "Atlantic/Reykjavik",
    "W. Europe Standard Time": "Europe/Berlin",
    "Central Europe Standard Time": "Europe/Budapest",
    "Romance Standard Time": "Europe/Paris",
    "Central European Standard Time": "Europe/Warsaw",
    "W. Central Africa Standard Time": "Africa/Lagos",
    "Jordan Standard Time": "Asia/Amman",
    "GTB Standard Time": "Europe/Bucharest",
    "Middle East Standard Time": "Asia/Beirut",
    "Egypt Standard Time": "Africa/Cairo",
    "E. Europe Standard Time": "Europe/Chisinau",
    "Syria Standard Time": "Asia/Damascus",
    "West Bank Standard Time": "Asia/Hebron",
    "South Africa Standard Time": "Africa/Johannesburg",
    "FLE Standard Time": "Europe/Kiev",
    "Israel Standard Time": "Asia/Jerusalem",
    "Kaliningrad Standard Time": "Europe/Kaliningrad",
    "Sudan Standard Time": "Africa/Khartoum",
    "Libya Standard Time": "Africa/Tripoli",
    "Namibia Standard Time": "Africa/Windhoek",
    "Arabic Standard Time": "Asia/Baghdad",
    "Turkey Standard Time": "Europe/Istanbul",
    "Arab Standard Time": "Asia/Riyadh",
    "Belarus Standard Time": "Europe/Minsk",
    "Russian Standard Time": "Europe/Moscow",
    "E. Africa Standard Time": "Africa/Nairobi",
    "Iran Standard Time": "Asia/Tehran",
    "Arabian Standard Time": "Asia/Dubai",
    "Astrakhan Standard Time": "Europe/Astrakhan",
    "Azerbaijan Standard Time": "Asia/Baku",
    "Russia Time Zone 3": "Europe/Samara",
    "Mauritius Standard Time": "Indian/Mauritius",
    "Saratov Standard Time": "Europe/Saratov",
    "Georgian Standard Time": "Asia/Tbilisi",
    "Volgograd Standard Time": "Europe/Volgograd",
    "Caucasus Standard Time": "Asia/Yerevan",
    "Afghanistan Standard Time": "Asia/Kabul",
    "West Asia Standard Time": "Asia/Tashkent",
    "Ekaterinburg Standard Time": "Asia/Yekaterinburg",
    "Pakistan Standard Time": "Asia/Karachi",
    "India Standard Time": "Asia/Calcutta",
    "Sri Lanka Standard Time": "Asia/Colombo",
    "Nepal Standard Time": "Asia/Kathmandu",
    "Central Asia Standard Time": "Asia/Almaty",
    "Bangladesh Standard Time": "Asia/Dhaka",
    "Omsk Standard Time": "Asia/Omsk",
    "Myanmar Standard Time": "Asia/Rangoon",
    "SE Asia Standard Time": "Asia/Bangkok",
    "Altai Standard Time": "Asia/Barnaul",
    "W. Mongolia Standard Time": "Asia/Hovd",
    "North Asia Standard Time": "Asia/Krasnoyarsk",
    "N. Central Asia Standard Time": "Asia/Novosibirsk",
    "Tomsk Standard Time": "Asia/Tomsk",
    "China Standard Time": "Asia/Shanghai",
    "North Asia East Standard Time": "Asia/Irkutsk",
    "Singapore Standard Time": "Asia/Singapore",
    "W. Australia Standard Time": "Australia/Perth",
    "Taipei Standard Time": "Asia/Taipei",
    "Ulaanbaatar Standard Time": "Asia/Ulaanbaatar",
    "North Korea Standard Time": "Asia/Pyongyang",
    "Aus Central W. Standard Time": "Australia/Eucla",
    "Transbaikal Standard Time": "Asia/Chita",
    "Tokyo Standard Time": "Asia/Tokyo",
    "Korea Standard Time": "Asia/Seoul",
    "Yakutsk Standard Time": "Asia/Yakutsk",
    "Cen. Australia Standard Time": "Australia/Adelaide",
    "AUS Central Standard Time": "Australia/Darwin",
    "E. Australia Standard Time": "Australia/Brisbane",
    "AUS Eastern Standard Time": "Australia/Sydney",
    "West Pacific Standard Time": "Pacific/Port_Moresby",
    "Tasmania Standard Time": "Australia/Hobart",
    "Vladivostok Standard Time": "Asia/Vladivostok",
    "Lord Howe Standard Time": "Australia/Lord_Howe",
    "Bougainville Standard Time": "Pacific/Bougainville",
    "Russia Time Zone 10": "Asia/Srednekolymsk",
    "Magadan Standard Time": "Asia/Magadan",
    "Norfolk Standard Time": "Pacific/Norfolk",
    "Sakhalin Standard Time": "Asia/Sakhalin",
    "Central Pacific Standard Time": "Pacific/Guadalcanal",
    "Russia Time Zone 11": "Asia/Kamchatka",
    "New Zealand Standard Time": "Pacific/Auckland",
    "UTC+12": "Etc/GMT-12",
    "Fiji Standard Time": "Pacific/Fiji",
    "Chatham Islands Standard Time": "Pacific/Chatham",
    "UTC+13": "Etc/GMT-13",
    "Tonga Standard Time": "Pacific/Tongatapu",
    "Samoa Standard Time": "Pacific/Apia",
    "Line Islands Standard Time": "Pacific/Kiritimati",
}



from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_linkedin
from arcade_linkedin.tools.share import create_text_post

rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_linkedin)


@tool_eval()
def linkedin_eval_suite():
    suite = EvalSuite(
        name="LinkedIn Tools Evaluation",
        system_message="You are an AI assistant with access to LinkedIn tools. Use them to help the user with their tasks.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Run code",
        user_message="post this transcription to linkedin. there may be some things that you need to clean up since it was spoken.: 'It is with great pleasure that I announce that I am now a member of the LinkedIn community! I'd like to thank the LinkedIn team for their support and encouragement in my journey to success. hash tag Y2K'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_text_post,
                args={
                    "text": "It is with great pleasure that I announce that I am now a member of the LinkedIn community! I'd like to thank the LinkedIn team for their support and encouragement in my journey to success. #Y2K",
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="text", weight=1.0),
        ],
    )

    return suite
